# -*- coding: utf-8 -*-
# ELEKTRONN - Neural Network Toolkit
#
# Copyright (c) 2014 - now
# Max-Planck-Institute for Medical Research, Heidelberg, Germany
# Authors: Marius Killinger, Gregor Urban

import numpy as np
import time, os, sys, gc
from matplotlib import pyplot as plt

import trainutils as ut
from warping import getWarpParams, warp2dJoint, warp3dJoint

#if __package__ is None:
#  sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__) ) ) )

try:
    import fadvise
    fadvise_avail = True
except:
    fadvise_avail = False

try:
    import malis
    malis_avail = True
except:
    malis_avail = False

###############################################################################################


def plotTrainingTarget(img, lab, stride=1):
    """
    Plots raw image vs label to check if valid batches are produced.
    Raw data is also shown overlaid with labels

    Parameters
    ----------

    img: 2d array
      raw image from batch
    lab: 2d array
      labels
    stride: int
      stride of labels

    """

    if len(lab) * stride != len(img):
        off = (len(img) - stride * len(lab)) // 2 // stride
        new_l = np.zeros((lab.shape[0] + 2 * off, lab.shape[1] + 2 * off))
        new_l[off:-off, off:-off] = lab
        lab = new_l

    plt.figure(figsize=(18, 6))
    plt.subplot(131)
    plt.imshow(img, interpolation='none', cmap=plt.get_cmap('gray'))
    plt.title('data')
    plt.subplot(132)
    plt.imshow(lab, interpolation='none', cmap=plt.get_cmap('gray'))
    plt.title('label')
    if img.shape == lab.shape:
        overlay = 0.75 * img + 0.25 * (1 - lab)
        plt.subplot(133)
        plt.imshow(overlay, interpolation='none', cmap=plt.get_cmap('gray'))
        plt.title('overlay')
    plt.show()
    return img - lab


def _transposeIgnoreCh(arr, T):
    """Transpose ``arr`` with permutation ``T`` but ignore channel axis with index 1"""
    T_ = np.copy(T)
    T_ += T_ > 0
    return np.transpose(arr, (T_[0], 1) + tuple(T_[1:]))


def _getValid3dTranspose(arr, ps):
    """For 3d ``arr`` and 2d patchsize ``ps`` returns list of valid transposes that cover the patch size"""
    all_T = [(0, 1, 2), (2, 0, 1), (1, 2, 0)]
    possible_T = []
    for T in all_T:
        sh = np.transpose(arr, T).shape
        if reduce(lambda x, y: (y[0] <= y[1]) and x, zip(ps, sh[:2]), True):  # check if first 2 dims are large enough
            possible_T.append(T)

    return possible_T


def _randomFlip(d, l, rng, aniso=True, upright_x=False):
    """Do joint random 90deg-rotation/mirroring on spatial axes for ``d`` and ``l``"""
    n_dim = len(d.shape) - 1
    if n_dim == 2:
        if not upright_x:
            if 0.5 < rng.rand():
                d = np.swapaxes(d, 1, 2)
                l = np.swapaxes(l, 0, 1)
            flip = rng.binomial(1, 0.5, 2) * 2 - 1  # gives strides of -1 and 1 for all spatial axis
            d = d[:, ::flip[0], ::flip[1]]  # this flips axis directions
            l = l[::flip[0], ::flip[1]]
        else:  # only flip the y-axis for upright images
            if 0.5 < rng.rand():
                d = d[:, :, ::-1]
                l = l[:, ::-1]

    elif n_dim == 3:
        if aniso:  # For anisotropic resolution only rotate in x-y-pane
            if 0.5 < rng.rand():
                d = np.swapaxes(d, 2, 3)
                l = np.swapaxes(l, 1, 2)
        else:
            T = rng.permutation(3)
            d = _transposeIgnoreCh(d, T)
            l = np.transpose(l, T)

        flip = rng.binomial(1, 0.5, 3) * 2 - 1
        d = d[::flip[0], :, ::flip[1], ::flip[2]]
        l = l[::flip[0], ::flip[1], ::flip[2]]

    return d, l


def _greyAugment(d, channels, rng):
    """
    Performs grey value (historgram) augmentations on ``d``. This is only applied to ``channels``
    (list of channels indices), ``rng`` is a random number generator
    """
    if channels == []:
        return d
    else:
        n_dim = len(d.shape) - 1
        k = len(channels)
        d = d.copy()  # d is still just a view, we don't want to change the original data so copy it
        alpha = 1 + (rng.rand(k) - 0.5) * 0.3  # ~ contrast
        c = (rng.rand(k) - 0.5) * 0.3  # mediates whether values are clipped for shadows or lights
        gamma = 2.0**(rng.rand(k) * 2 - 1)  # sample from [0.5,2] with mean 0
        if n_dim == 3:
            d[:, channels] = d[:, channels] * alpha[None, :, None, None] + c[None, :, None, None]
            d[:, channels] = np.clip(d[:, channels], 0, 1)  # clip to valid range (otherwise the above has little effect)
            d[:, channels] = d[:, channels]**gamma[None, :, None, None]
        elif n_dim == 2:
            d[channels] = d[channels] * alpha[:, None, None] + c[:, None, None]
            d[channels] = np.clip(d[channels], 0, 1)
            d[channels] = d[channels]**gamma[:, None, None]
    return d


def _stripCubes(d, l, off, ldtype):
    """
    Determine minimal trainable image data size (depending on the size of labels and CNN offset).
    It is assumed that labels lie in the center of the corresponding raw image cube.
    Finally d and l have same shapes in x,y,z, where l is zero-padded or d is cut.
    """
    if len(off) == 2:
        off = np.concatenate(([0, ], off))

    d_off = []
    l_cut = []
    for i, (d_sh, l_sh) in enumerate(zip(
            np.array(d.shape)[[0, 2, 3]], l.shape)):
        assert ((d_sh - l_sh) - 2 * off[i]) % 2 == 0
        x = ((d_sh - l_sh) - 2 * off[i]) / 2  #  This the offset with which data needs to be cut
        # The missing length of the labels, but data must still be 2*off[i] larger, by two (one cut from each side)
        if x < 0:  # More labels than data
            l_cut.append(-x)  # These are the superfluous label stripes
            x = 0
        else:  # More data than labels
            l_cut.append(0)
        d_off.append(x)  # These are the superfluous data stripes

    # cut minimal trainable data
    d = d[d_off[0]:d.shape[0] - d_off[0], :, d_off[1]:d.shape[2] - d_off[1], d_off[2]:d.shape[3] - d_off[2]]

    # cut minimal trainable labels
    if np.any(l_cut):
        print "Warning: the CNN offset is so large that labels are cut off!\nOn each side of the cube %s" % (l_cut)
    new_l_sh = np.array(d.shape)[[0, 2, 3]]  #map(lambda x: x[0] - 2*x[1] , zip(np.array(d.shape)[[0,2,3]], off))
    new_l = np.zeros(new_l_sh, dtype=ldtype)
    l = l[l_cut[0]:l.shape[0] - l_cut[0], l_cut[1]:l.shape[1] - l_cut[1], l_cut[2]:l.shape[2] - l_cut[2]]

    new_l[off[0]:new_l.shape[0] - off[0], off[1]:new_l.shape[1] - off[1], off[2]:new_l.shape[2] - off[2]] = l

    assert d.shape[2:] == new_l.shape[1:]
    assert d.shape[0] == new_l.shape[0]
    return d, new_l


def _borderTreatment(data_list, ps, border_mode, n_dim):
    def treatArray(data):
        if border_mode == 'keep':
            return data

        if n_dim == 3:
            sh = (data.shape[0], ) + data.shape[2:]  # exclude channel (z,x,y)
        else:
            sh = data.shape[2:]  # (x,y)

        if border_mode == 'crop':
            excess = map(lambda x: int((x[0] - x[1]) // 2), zip(sh, ps))
            if n_dim == 3:
                data = data[excess[0]:excess[0] + ps[0], :, excess[1]:excess[1] + ps[1], excess[2]:excess[2] + ps[2]]
            elif n_dim == 2:
                data = data[:, :, excess[0]:excess[0] + ps[0], excess[1]: excess[1] + ps[1]]

        else:
            excess_l = map(lambda x: int(np.ceil(float(x[0] - x[1]) / 2)), zip(ps, sh))
            excess_r = map(lambda x: int(np.floor(float(x[0] - x[1]) / 2)), zip(ps, sh))
            if n_dim == 3:
                pad_with = [(excess_l[0], excess_r[0]), (0, 0), (excess_l[1], excess_r[1]), (excess_l[2], excess_r[2])]
            else:
                pad_with = [(0, 0), (0, 0), (excess_l[0], excess_r[0]), (excess_l[1], excess_r[1])]

            if border_mode == 'mirror':
                data = np.pad(data, pad_with, mode='symmetric')

            if border_mode == '0-pad':
                data = np.pad(data, pad_with, mode='constant', constant_values=0)

        return data

    return [treatArray(d) for d in data_list]


def _make_affinities(labels, nhood=None, size_thresh=1):
    """
    Construct an affinity graph from a segmentation (IDs) 
    
    Segments with ID 0 are regarded as disconnected
    The spatial shape of the affinity graph is the same as of seg_gt.
    This means that some edges are are undefined and therefore treated as disconnected.
    If the offsets in nhood are positive, the edges with largest spatial index are undefined.
    
    Connected components is run on the affgraph to relabel the IDs locally.
    
    Parameters
    ----------
    
    labels: 4d np.ndarray, int (any precision)
        Volumes of segmentation IDs (bs, z, x, y)
    nhood: 2d np.ndarray, int
        Neighbourhood pattern specifying the edges in the affinity graph
        Shape: (#edges, ndim)
        nhood[i] contains the displacement coordinates of edge i
        The number and order of edges is arbitrary
    size_thresh: int
        Size filters for connected compontens, smaller objects are mapped to BG
        
        
    Returns
    -------
    
    aff: 5d np.ndarray int16
        Affinity graph of shape (bs, #edges, x, y, z)
        1: connected, 0: disconnected  
    seg_gt:
        4d np.ndarray int16
        Affinity graph of shape (bs, x, y, z)
        Relabelling of components     
    """

    if not malis_avail:
        raise RuntimeError("Please install malis to use affinities")

    if nhood is None:
        nhood = np.eye(3, dtype=np.int32)

    aff_sh = [labels.shape[0], nhood.shape[0], ] + list(labels.shape[1:])
    out_aff = np.zeros(aff_sh, dtype=np.int16)
    out_seg = np.zeros(labels.shape, dtype=np.int16)
    for i, l in enumerate(labels):
        out_aff[i] = malis.seg_to_affgraph(l, nhood)
        # we throw away the seg sizes
        out_seg[i], _ = malis.affgraph_to_seg(out_aff[i], nhood, size_thresh)
    return out_aff, out_seg


class CNNData(object):
    """
    Patch creation and data handling interface for image like training data

    Parameters
    ----------

    patch_size: 2/3-tuple
      Specifying CNN input shape of a single example, **without** channels: (x,y)/(z,x,y)
    stride: 2/3-tuple
      Specifying CNN output stride. May be ``None`` for scalar labels
    offset: 2/3-tuple
      Specifying overall CNN convolution border. May be ``None`` for scalar labels
    n_dim: int
      2 or 3, CNN dimension
    n_lab: int
      Number of distinct classes/labels, if not provided (->None) this is automatically inferred (slow!)
    anistropic_data: Bool
      If True 2d slices are only cut and rotated along z-axis, otherwise all 3 alignments are used
    mode: str
      Mode that describes the kind of data and labels: img-img or img-scalar. If the labels are
      scalar but the data is a stack (along z-axis) of many examples, the many scalar labels should be
      stacked to a vector. For vect-scalar training use the ``TrainData``-class instead.
    zchxy_order: Bool
      If the data files are already in memory layout (z,ch,x,y)/(z,x,y), this option must be set to True,
      which makes data loading faster.
    border_mode: string
      For img-scalar training: specifies how to treat images that don't match a valid CNN input size
    upright_x: Bool
      If ``True``, image augmentation leaves the upright position of natural images intact, e.g. they are
      only mirrored horizontally, not vertically. Note: the horizontal direction corresponds to 'y' (because the
      'x' comes before 'y' and the vertical comes before horizontal in numpy)!
    float_label: Bool
       Whether to return labels as float32 (for regression) or int16 (for classification)
    affinity: str/False
       False/'affinity'/'malis': malis returs additionally the segmentation IDs
    """

    def __init__(self,
                 patch_size=None,
                 stride=None,
                 offset=None,
                 n_dim=2,
                 n_lab=None,
                 anistropic_data=False,
                 mode='img-img',
                 zchxy_order=False,
                 border_mode='crop',
                 pre_process=None,
                 upright_x=False,
                 float_label=False,
                 affinity=False):
        print '\n'
        self.n_dim = n_dim
        if not hasattr(patch_size, '__len__'):
            patch_size = (patch_size, ) * n_dim
        else:
            assert len(patch_size) == n_dim
        if not hasattr(stride, '__len__') and stride is not None:
            stride = (stride, ) * n_dim
        if not hasattr(offset, '__len__') and offset is not None:
            offset = (offset, ) * n_dim

        self.patch_size = np.array(patch_size, dtype=np.int)
        if mode == 'img-img':
            self.stride = np.array(stride, dtype=np.int)
            self.offset = np.array(offset, dtype=np.int)

        self.n_lab = n_lab
        self.aniso = anistropic_data
        self.mode = mode
        self.zchxy_order = zchxy_order
        self.border_mode = border_mode
        self.pre_process = pre_process
        self.upright_x = upright_x
        self.ldtype = np.float32 if float_label else np.int16
        self.affinity = affinity
        if n_dim == 3 and upright_x:
            print "Warning: the data is 3-dimensional and the 'upright_x'-flag is active, but it works only for 2d!"

        self.rng = np.random.RandomState(np.uint32((time.time() * 0.0001 - int(time.time() * 0.0001)) * 4294967295))
        self.pid = os.getpid()
        self.gc_count = 1

        self._sampling_weight = None
        self._training_count = None
        self._valid_count = None
        self.n_successful_warp = 0
        self.n_failed_warp = 0

        # Actual data
        self.names = []
        self.info = []

        self.valid_d = []
        self.valid_l = []
        self.valid_i = []

        self.train_d = []
        self.train_l = []
        self.train_i = []

    def __repr__(self):
        return "%i-class Data Set with %i input channel(s):\n#train cubes: %i and #valid cubes: %i" \
                % (self.n_lab, self.n_ch, self._training_count, self._valid_count)

    def addDataFromFile(self,
                        d_path,
                        l_path,
                        d_files,
                        l_files,
                        cube_prios=None,
                        valid_cubes=[],
                        downsample_xy=False):
        """
        Parameters
        ----------

        d_path/l_path: string
          Directories to load data from
        d_files/l_files: list
          List of data/label files in <path> directory (must be in the same order!). Each list
        element is a tuple in the form **(<Name of h5-file>, <Key of h5-dataset>)**
        cube_prios: list
          (not normalised) list of sampling weights to draw examples from the respective cubes.
          If None the cube sizes are taken as priorities.
        valid_cubes: list
          List of indices for cubes (from the file-lists) to use as validation data and exclude from training,
          may be empty list to skip performance estimation on validation data.
        """
        self.names += d_files
        if fadvise_avail:
            names = reduce(lambda x, y: x + [d_path + y[0][0], l_path + y[1][0]], zip(d_files, l_files), [])
            fadvise.willneed(names)
        # returns lists of cubes, info is a tuple per cube
        data, label, info = self._read_images(d_path, l_path, d_files, l_files, downsample_xy)

        if self.mode == 'img-scalar':
            data = _borderTreatment(data, self.patch_size, self.border_mode, self.n_dim)

        if self.pre_process:
            if self.pre_process == 'standardise':
                M = np.mean(map(np.mean, data))
                S = np.mean(map(np.std, data))
                data = map(lambda x: (x - M) / S, data)
                print "Data is standardised. Original mean: %.g, original std %.g" % (M, S)
                self.data_mean = M
                self.data_std = S

            else:
                raise NotImplementedError("Pre-processing %s is not implemented" % self.pre_process)

        if self.n_lab is None:
            unique = [np.unique(l) for l in label]
            self.n_lab = np.unique(np.hstack(unique)).size

        default_info = (np.ones(self.n_lab), np.zeros(self.n_lab))
        info = map(lambda x: default_info if x is None else x, info)
        self.info += info

        prios = []
        # Distribute Cubes into training and valid list
        for k, (d, l, i) in enumerate(zip(data, label, info)):
            if k in valid_cubes:
                self.valid_d.append(d)
                self.valid_l.append(l)
                self.valid_i.append(i)
            else:
                self.train_d.append(d)
                self.train_l.append(l)
                self.train_i.append(i)
                # If no priorities are given: sample proportional to cube size
                prios.append(l.size)

        if cube_prios is None or cube_prios == []:
            prios = np.array(prios, dtype=np.float)
        else:  # If priorities are given: sample irrespective of cube size
            prios = np.array(cube_prios, dtype=np.float)

            # sample example i if: batch_prob[i] < p
        self._sampling_weight = np.hstack((0, np.cumsum(prios / prios.sum())))
        self._training_count = len(self.train_d)
        self._valid_count = len(self.valid_d)

        print self.__repr__()
        print '\n'

    def addDataFromNdarray(self,
                           d_train,
                           l_train,
                           d_valid=[],
                           l_valid=[],
                           cube_prios=None):
        """
        Parameters
        ----------

        d_train: list of numpy arrays
          the input data for Training
        l_train: list of numpy arrays
          the labels     for Training
        d_valid: list of numpy arrays
          the input data for validation [OPTIONAL]
        l_valid: list of numpy arrays
          the labels     for validation [OPTIONAL]
        cube_prios: list of floats
          Default: None --> probability of sampling Training data from a cube is proportional to its size
        """
        if type(d_train) is not type([]):
            d_train = [d_train]
            l_train = [l_train]
        self.names = ["directly_added_data_" + str(i) for i in range(len(d_train))]
        if self.n_lab is None:
            unique = [np.unique(l) for l in l_train]
            self.n_lab = np.unique(np.hstack(unique)).size
        default_info = (np.ones(self.n_lab), np.zeros(self.n_lab))
        info = map(lambda x: default_info if x is None else x, default_info)
        self.info = info

        if len(d_train) and d_train[0].n_dim == 3 or len(d_valid) and d_valid[0].n_dim == 3:
            self.n_ch = 1
        else:
            self.n_ch = d_train[0].shape[0] if len(d_train) else d_valid[0].shape[0]

        if self.mode == 'img-scalar':
            self.valid_d += _borderTreatment(d_valid, self.patch_size, self.border_mode, self.n_dim)
            self.train_d += _borderTreatment(d_train, self.patch_size, self.border_mode, self.n_dim)
        else:
            self.valid_d += d_valid
            self.train_d += d_train

        self.valid_l += l_valid
        self.train_l += l_train

        self.train_i = info
        self.valid_i = info

        if cube_prios is not None:
            prios = np.array(cube_prios, dtype=np.float)
        else:
            prios = np.array([l.size for l in self.train_l], dtype=np.float)

        prios = np.array(prios, dtype=np.float)
        # sample example i if: batch_prob[i] < p
        self._sampling_weight = np.hstack((0, np.cumsum(prios / prios.sum())))
        self._training_count = len(self.train_d)
        self._valid_count = len(self.valid_d)

        print self.__repr__()

    def _allocBatch(self, batch_size):
        patch_size = self.patch_size
        if self.n_dim == 2:

            images = np.zeros((batch_size, self.n_ch, patch_size[0], patch_size[1]), dtype='float32')
            if self.mode == 'img-img':
                off = self.offset
                label = np.zeros((batch_size, patch_size[0] - 2 * off[0],
                                 patch_size[1] - 2 * off[1]),
                                 dtype=self.ldtype)
            else:
                label = np.zeros((batch_size, ), dtype=self.ldtype)

        elif self.n_dim == 3:
            images = np.zeros((batch_size, patch_size[0], self.n_ch, patch_size[1],
                              patch_size[2]),
                              dtype='float32')
            if self.mode == 'img-img':
                off = self.offset
                label = np.zeros((batch_size, patch_size[0] - 2 * off[0],
                                  patch_size[1] - 2 * off[1], patch_size[2] - 2 * off[2]),
                                  dtype=self.ldtype)
            else:
                label = np.zeros((batch_size, ), dtype=self.ldtype)

        return images, label

    def getbatch(self,
                 batch_size=1,
                 source='train',
                 strided=True,
                 flip=True,
                 grey_augment_channels=[],
                 ret_info=False,
                 warp_on=False,
                 ignore_thresh=0.0,
                 ret_example_weights=False):
        """
        Prepares a batch by randomly sampling, shifting and augmenting patches from the data

        Parameters
        ----------
        batch_size: int
          Number of examples in batch (for CNNs often just 1)
        source: string
          Data set to draw data from: 'train'/'valid'
        strided: Bool
          If True the labels are sub-sampled according to the CNN output stride.
          Non-strided labels requires MFP in the CNN!
        flip: Bool
          If True examples are mirrored and rotated by 90 deg randomly
        grey_augment_channels: list
          List of channel indices to apply grey-value augmentation to
        ret_info: Bool
          If True additional information for reach batch example is returned. Currently implemented are two info
          arrays to indicate the labelling mode. The first dimension of those arrays is the batch_size!
        warp_on: Bool/Float(0,1)
          Whether warping/distortion augmentations are applied to examples (slow --> use multiprocessing)
          If this is a float number, warping is applied to this fraction of examples e.g. 0.5 --> every other example
        ignore_thresh: float
          If the fraction of negative labels in an example patch exceeds this threshold, this example is discarded
          (Negative labels are ignored for training [but could be used for unsupervised label propagation]).

        Returns
        -------

        data:
          [bs, ch, x, y] or [bs, z, ch, x, y] for 2d and 3d CNNS

        label:
          [bs, x, y] or [bs, z, x, y]
        info1:
          (optional) [bs, n_lab]
        info2:
          (optional) [bs, n_lab]
        """
        # This is especially required for multiprocessing
        self._reseed()
        images, label = self._allocBatch(batch_size)
        infos = []
        patch_count = 0
        while patch_count < batch_size:  # Loop to fill up batch with examples
            d, l, info = self._getcube(source)  # get cube randomly
            d, l = self._warpAugment(d, l, warp_on, ignore_thresh, self.upright_x)  # doesn't change l if l is non-image

            if (ignore_thresh != 0.0) and (not np.any(info[1])) and (float(np.count_nonzero([l < 0])) / l.size) > ignore_thresh:
                continue  # do not use cubes which have no information

            if flip:
                if self.patch_size[-1] != self.patch_size[-2]:
                    raise RuntimeError("Cannot apply 'flip' to image if x and y have different sizes")

                if self.mode == 'img-img':
                    d, l = _randomFlip(d, l, self.rng, self.aniso, self.upright_x)
                else:
                    d, _ = _randomFlip(d, d, self.rng, self.aniso, self.upright_x)  # lazy hack to exclude labels from transform

            if source == "train":  # no grey augmentation for testing
                d = _greyAugment(d, grey_augment_channels, self.rng)

            label[patch_count] = l
            images[patch_count] = d
            infos.append(info)
            patch_count += 1

        if ret_example_weights:
            weights = self.getExampleWeights(images, label)

        if strided:
            label = self._stridedLabels(label)

        ret = [images, label]

        if self.affinity == 'malis':
            aff, seg = _make_affinities(label)  # [bs, z, x, y, 3]
            ret = [images, aff, seg]

        if self.affinity == 'affinity':
            aff, seg = _make_affinities(label)  # [bs, z, x, y, 3]
            ret = [images, aff]

        if ret_info:  # info is now a list(bs) of tuples(2)
            infos = np.atleast_3d(np.array(infos, dtype=np.int16))  # (bs, 2, 5)
            info1 = infos[:, 0]
            info2 = infos[:, 1]
            ret += [info1, info2]

        if ret_example_weights:
            ret += [weights, ]

        self.gc_count += 1
        if self.gc_count % 1000 == 0:
            gc.collect()

        return ret

    def _warpAugment(self, d, l, warp_on, ignore_thresh, upright_x):
        if (warp_on is True) or (warp_on == 1):  # always warp
            do_warp = True
        elif (0 < warp_on < 1):  # warp only a fraction of examples
            do_warp = True if (self.rng.rand() < warp_on) else False
        else:  # never warp
            do_warp = False

        if False and upright_x and do_warp:
            assert self.n_dim == 2, "upright_x only works for 2d"
            raise NotImplementedError()
        elif do_warp:
            ext_size, rot, shear, scale, stretch, twist = getWarpParams(self.patch_size)
            try:
                d, l = self._cutPatch(d, l, ps=ext_size, thresh=ignore_thresh)
                if self.n_dim == 2:
                    d, l = warp2dJoint(d, l, self.patch_size, rot, shear, scale, stretch)  # ignores label if non-image
                else:
                    d, l = warp3dJoint(d, l, self.patch_size, rot, shear, scale, stretch, twist)  # ignores label if non-image
                self.n_successful_warp += 1

            except ValueError:  # the ext_size is to big for this data cube
                self.n_failed_warp += 1
                d, l = self._cutPatch(d, l, thresh=ignore_thresh)  # Don't do warping
        else:  # do not warp
            d, l = self._cutPatch(d, l, thresh=ignore_thresh)  # Don't do warping

        return d, l

    def getExampleWeights(self, raw_rec, lab, gain=2.0, blurr=False):
        off = self.offset
        previous_pred = raw_rec[:, :, -1]  # the last channel is the prediction
        previous_pred = previous_pred[:, off[0]:-off[0], off[1]:-off[1], off[2]:-off[2]]
        if blurr:
            lab = lab
            raise NotImplementedError()

        lab = self._stridedLabels(lab)
        previous_pred = self._stridedLabels(previous_pred)
        diff = lab - previous_pred  # misses are (+) and clutter is (-)
        weights = np.ones_like(lab, dtype=np.float32)
        weights += (diff < -0.1) * gain * (-diff - 0.1)
        weights += (diff > 0.05) * gain * (diff - 0.05)
        # This gives ca. 1.5 times weight, so maybe the same LR can be used
        return weights

    def _getcube(self, source):
        """Draw an example cube according to sampling weight on training data, or randomly on valid data"""
        if source == 'train':
            p = self.rng.rand()
            i = np.flatnonzero(self._sampling_weight <= p)[-1]
            d, l, info = self.train_d[i], self.train_l[i], self.train_i[i]
        elif source == "valid":
            if len(self.valid_d) == 0:
                print "Validation Set empty. Disable testing on validation set."
                d, l, info = [], [], []

            i = self.rng.randint(0, len(self.valid_d))
            d = self.valid_d[i]
            l = self.valid_l[i]
            info = self.valid_i[i]

        else:
            raise ValueError("Unkonown data source")

        return d, l, info

    def _cutPatch(self, img, lab, ps=None, thresh=0, it=0):
        """
        Cut a patch from a cube of data and label
        To enable deformations the patch must be cut with ``ps`` shape. ``thresh`` specifies threshold on negative
        (i.e. ignore) labels, then another iteration ``it`` is started (but maximal 10)
        """
        if ps is None:
            ps = self.patch_size

        if self.n_dim == 3:
            try:
                shift = [int(self.rng.randint(0, s - p, 1)) for p, s in zip(ps, np.array(img.shape)[[0, 2, 3]])]
            except ValueError:
                if np.all(np.equal(ps, np.array(img.shape)[[0, 2, 3]])):
                    shift = [0, 0, 0]
                else:
                    raise ValueError("Image smaller than patch size: Image shape=%s, patch size=%s"
                                     % (img.shape[1:], ps))

            cut_img = img[shift[0]:shift[0] + ps[0], :, shift[1]:shift[1] + ps[1], shift[2]:shift[2] + ps[2]]
            if self.mode == 'img-img':
                off = self.offset
                cut_lab = lab[off[0] + shift[0]:shift[0] + ps[0] - off[0],
                              off[1] + shift[1]:shift[1] + ps[1] - off[1],
                              off[2] + shift[2]: shift[2] + ps[2] - off[2]]
            else:
                cut_lab = lab

        else:  # 2d
            if self.aniso or self.mode == 'img-scalar':  # no transposition of axes for anisotropic data
                imgT, labT = img, lab

            else:  # For isotropic data slices must not exclusively be cut in z direction
                # cut also perpendicular to x or y axis, this is not flipping (see separate function _flip)!
                possible_T = _getValid3dTranspose(lab, ps)
                i = self.rng.randint(0, len(possible_T))
                imgT = _transposeIgnoreCh(img, possible_T[i])  # img has one dim more
                labT = np.transpose(lab, possible_T[i])

            z_pos = self.rng.randint(0, imgT.shape[0])
            try:
                shift = [int(self.rng.randint(0, s - p, 1)) for p, s in zip(ps, imgT.shape[-2:])]
            except ValueError:
                if np.all(np.equal(ps, imgT.shape[-2:])):
                    shift = [0, 0]
                else:
                    raise ValueError("Image smaller than patch size: Image shape=%s, patch size=%s"
                                     % (imgT.shape[1:], ps))

            cut_img = imgT[z_pos, :, shift[0]:shift[0] + ps[0], shift[1]:shift[1] + ps[1]]

            if self.mode == 'img-img':
                off = self.offset
                cut_lab = labT[z_pos, off[0] + shift[0]:shift[0] + ps[0] - off[0],
                               off[1] + shift[1]:shift[1] + ps[1] - off[1]]
            else:
                cut_lab = labT[z_pos]

                # check if there are enough non-ignore (i.e. positive) labels , but MAX 10 time, then use current result
        if (it < 10) and (
                thresh != 0.0) and (float(np.count_nonzero([cut_lab < 0])) / cut_lab.size) > thresh:
            return self._cutPatch(img, lab, ps, thresh, it + 1)

        else:
            return cut_img, cut_lab

    def _reseed(self):
        """Reseeds the rng if the process ID has changed!"""
        current_pid = os.getpid()
        if current_pid != self.pid:
            self.pid = current_pid
            self.rng.seed(np.uint32((time.time() * 0.0001 - int(time.time() * 0.0001)) * 4294967295 + self.pid))
            print "Reseeding RNG in Process with PID:", self.pid

    def _stridedLabels(self, lab):
        if self.n_dim == 3:
            return lab[:, ::self.stride[0], ::self.stride[1], ::self.stride[2]]
        else:
            return lab[:, ::self.stride[0], ::self.stride[1]]

    def _read_images(self, d_path, l_path, d_files, l_files, downsample_xy):
        """
        Image files on disk are expected to be in order (ch,x,y,z) or (x,y,z)
        But image stacks are returned as (z,ch,x,y) and label as (z,x,y,) irrespective of the order in the file.
        If the image files have no channel this dimension is extended to a singleton dimension.
        """
        data, label, info = [], [], []
        if len(d_files) != len(l_files):
            raise ValueError("d_files and l_files must be lists of same length!")
        for (d_f, d_key), (l_f, l_key) in zip(d_files, l_files):
            print 'Loading %s' % d_f,
            d = ut.h5Load(d_path + d_f, d_key)
            print 'Loading %s' % l_f
            l = ut.h5Load(l_path + l_f, l_key)
            try:
                info_1 = ut.h5Load(l_path + l_f, 'info')
                info.append(info_1)
            except KeyError:
                info.append(None)

            if not self.zchxy_order:
                if len(d.shape) == 4:
                    self.n_ch = d.shape[0]
                    print "Data has %i channels" % self.n_ch
                elif len(d.shape) == 3:  # We have no channels in data
                    self.n_ch = 1
                    d = d[None, :, :, :]  # add (empty) 0-axis

                if l.size == 0:
                    l = np.zeros_like(d[0], dtype=self.ldtype)
                elif self.mode == 'img-scalar':
                    assert len(l.shape) == 1, "Scalar labels must be 1d"

    # Transpose such that access is optimal
                d = np.transpose(d, (3, 0, 1, 2))  # (ch,x,y,z)-->(z,ch,x,y)
                if self.mode == 'img-img':
                    l = np.transpose(l, (2, 0, 1))  #    (x,y,z)-->(z,x,y)
                    d, l = _stripCubes(d, l, self.offset, self.ldtype)

            else:  # data in memory layout:
                if len(d.shape) == 4:
                    self.n_ch = d.shape[1]
                    print "Data has %i channels" % self.n_ch
                elif len(d.shape) == 3:  # We have no channels in data
                    self.n_ch = 1
                    d = d[:, None, :, :]  # add (empty) 0-axis

                if l.size == 0:
                    sh = (d.shape[0], ) + d.shape[2:]
                    l = np.zeros_like(sh, dtype=self.ldtype)
                elif self.mode == 'img-scalar':
                    assert len(l.shape) == 1, "Scalar labels must be 1d"

                if self.mode == 'img-img':
                    d, l = _stripCubes(d, l, self.offset, self.ldtype)

            # determine normalisation depending on int or float type
            if d.dtype in [np.int, np.int8, np.int16, np.int32, np.uint32,
                           np.uint, np.uint8, np.uint16, np.uint32, np.uint32]:
                m = 255
            else:
                m = 1

            d = np.ascontiguousarray(d, dtype=np.float32) / m
            if (self.ldtype is not l.dtype):
                m = l.max()
                M = np.iinfo(self.ldtype).max
                if m > M:
                    raise ValueError("Loading of data: labels must be cast to %s, but %s cannot store value %g, maximum allowed value: %g. You may try to renumber labels."
                                     % (self.ldtype, self.ldtype, m, M))

            l = np.ascontiguousarray(l, dtype=self.ldtype)

            if downsample_xy:
                f = int(downsample_xy)
                l_sh = l.shape
                cut = np.mod(l_sh, f)

                d = d[:, :, :l_sh[-2] - cut[-2], :l_sh[-1] - cut[-1]]
                sh = d[:, :, ::f, ::f].shape
                new_d = np.zeros(sh, dtype=np.float32)

                l = l[:, :l_sh[-2] - cut[-2], :l_sh[-1] - cut[-1]]
                sh = l[:, ::f, ::f].shape
                new_l = np.zeros(sh, dtype=self.ldtype)

                for i in xrange(f):
                    for j in xrange(f):
                        new_d += d[:, :, i::f, j::f]
                        new_l += l[:, i::f, j::f]

                d = new_d / f**2
                l = new_l / f**2

            gc.collect()

            print "Internal data.shape=%s, label.shape=%s" % (d.shape, l.shape)
            print '---'
            data.append(d)
            label.append(l)

        return data, label, info

    ##############################################################################################################


if __name__ == "__main__":
    print "Testing CNNData"
    if __package__ is None:
        sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

    from elektronn.net import netutils
    from parallelisation import BackgroundProc

    _data_path = os.path.expanduser('~/devel/data/BirdGT/')  # (*) Path to data dir
    # _data_path        = os.path.expanduser('~/mnt_ssh/home/mkilling/devel/data/BirdGT/') # (*) Path to data dir
    _label_path = _data_path
    _d_files = [('raw_cube0-crop.h5', 'raw')]
    _l_files = [('cube0_barrier-int16.h5', 'labels')]

    cube_prios = None
    n_lab = 2  # int or None for auto

    n_dim = 3
    desired_input = [
        80, 300, 300
    ]  # (*) <int> or <2/3-tuple> in (x,y)/(x,y,z)-order for anisotropic CNN
    filters = [6, 4, 4, 4, 4, 4]  # [1,1,1]
    pool = [2, 2, 1, 1, 1, 1]  # [1,1,1]
    MFP = False

    grey_augment_channels = [0]  # List of channel indices to apply transform
    flip_data = True

    dimensions = netutils.CNNCalculator(filters,
                                        pool,
                                        desired_input,
                                        MFP=MFP,
                                        force_center=False,
                                        n_dim=n_dim)
    print dimensions
    patch_size = dimensions.input
    off = int(dimensions.offset[0])

    D = CNNData(patch_size,
                dimensions.pred_stride,
                dimensions.offset,
                n_dim,
                n_lab=n_lab,
                anistropic_data=True,
                zchxy_order=False,
                mode='img-img',
                border_mode='keep')
    D.addDataFromFile(_data_path,
                      _label_path,
                      _d_files,
                      _l_files,
                      cube_prios=cube_prios,
                      valid_cubes=[])

    #----------------------------------------------------------------------------

    kwargs = dict(batch_size=1,
                  strided=False,
                  flip=flip_data,
                  grey_augment_channels=grey_augment_channels,
                  ret_info=True,
                  warp_on=True,
                  ignore_thresh=0.5)

    data, label, info1, info2 = D.getbatch(
        **kwargs)  # (bs, z, ch, x, y), (bs, z, x, y)

    img = data[0, :, 0]
    for i in xrange(img.shape[0]):
        plt.imsave('/tmp/%i-img.png' % (i), img[i, ], cmap='gray')

    lab = label[0]
    for i in xrange(lab.shape[0]):
        plt.imsave('/tmp/%i-lab.png' % (i + off), lab[i])



# -*- coding: utf-8 -*-
# ELEKTRONN - Neural Network Toolkit
#
# Copyright (c) 2014 - now
# Max-Planck-Institute for Medical Research, Heidelberg, Germany
# Authors: Marius Killinger, Gregor Urban

import numpy as np

import urllib2
import cPickle
import os, time, re

try:
    sklearn_avail = True
    from sklearn import cross_validation
except:
    sklearn_avail = False

import warping
import trainutils as ut


def sort_human(file_names):
    """ Sort the given list in the way that humans expect."""
    convert = lambda text: int(text) if text.isdigit() else text
    alphanum_key = lambda key: [convert(c) for c in re.split('([0-9]+)', key)]
    file_names.sort(key=alphanum_key)


class Data(object):
    """
    Load and prepare data, Base-Obj
    """

    def __init__(self, n_lab=None):
        self._pos = 0
        if isinstance(self.train_d, np.ndarray):
            self._training_count = self.train_d.shape[0]
            if n_lab is None:
                self.n_lab = np.unique(self.train_l).size
            else:
                self.n_lab = n_lab
        elif isinstance(self.train_d, list):
            self._training_count = len(self.train_d)
            if n_lab is None:
                unique = [np.unique(l) for l in self.train_l]
                self.n_lab = np.unique(np.hstack(unique)).size
            else:
                self.n_lab = n_lab

        self.example_shape = self.train_d[0].shape
        self.n_ch = self.example_shape[0]

        self.rng = np.random.RandomState(np.uint32((time.time() * 0.0001 - int(time.time() * 0.0001)) * 4294967295))
        self.pid = os.getpid()
        print self.__repr__()
        self._perm = self.rng.permutation(self._training_count)

    def _reseed(self):
        """Reseeds the rng if the process ID has changed!"""
        current_pid = os.getpid()
        if current_pid != self.pid:
            self.pid = current_pid
            self.rng.seed(np.uint32((time.time() * 0.0001 - int(time.time() * 0.0001)) * 4294967295 + self.pid))
            print "Reseeding RNG in Process with PID:", self.pid

    def __repr__(self):
        return "%i-class Data Set: #training examples: %i and #validing: %i" \
        % (self.n_lab, self._training_count, len(self.valid_d))

    def getbatch(self, batch_size, source='train'):
        if source == 'train':
            if (self._pos + batch_size) < self._training_count:
                self._pos += batch_size
                slice = self._perm[self._pos - batch_size:self._pos]
            else:  # get new permutation
                self._perm = self.rng.permutation(self._training_count)
                self._pos = 0
                slice = self._perm[:batch_size]

            if isinstance(self.train_d, np.ndarray):
                return (self.train_d[slice], self.train_l[slice])

            elif isinstance(self.train_d, list):
                data = np.array([self.train_d[i] for i in slice])
                label = np.array([self.train_l[i] for i in slice])
                return (data, label)

        elif source == 'valid':
            data = self.valid_d[:batch_size]
            label = self.valid_l[:batch_size]
            return (data, label)

        elif source == 'test':
            data = self.test_d[:batch_size]
            label = self.test_l[:batch_size]
            return (data, label)

    def createSplitPerm(self, size, subset_ratio=0.8, seed=None):
        rng = np.random.RandomState(np.uint32((time.time())))
        if seed is not None:
            rng.seed(np.uint32(seed))
        perm = rng.permutation(size)
        k = int(size * subset_ratio)
        return perm[:k]

    def createCVSplit(self, data, label, n_folds=3, use_fold=2, shuffle=False, random_state=None):
        if not sklearn_avail:
            raise RuntimeError("Please install sklearn to create CV splits")

        cv = cross_validation.KFold(len(data), n_folds, shuffle=shuffle, random_state=random_state)
        for fold, (train_i, valid_i) in enumerate(cv):
            if fold == use_fold:
                self.valid_d = data[valid_i]
                self.valid_l = label[valid_i]
                self.train_d = data[train_i]
                self.train_l = label[train_i]

    def splitData(self, data, label, valid_size, split_no=0):
        rng = np.random.RandomState(split_no)
        perm = rng.permutation(len(data))
        self.valid_d = data[perm[:valid_size]]
        self.valid_l = label[perm[:valid_size]]
        self.train_d = data[perm[valid_size:]]
        self.train_l = label[perm[valid_size:]]

    def makeExampleSubset(self, subset_ratio=0.8, seed=None):
        ix = self.createSplitPerm(len(self.train_d), subset_ratio, seed)
        self.train_d = self.train_d[ix]
        self.train_l = self.train_l[ix]
        self._training_count = len(self.train_d)
        self._perm = self.rng.permutation(self._training_count)

    def makeFeatureSubset(self, subset_ratio=0.8, seed=None):
        ix = self.createSplitPerm(self.train_d.shape[1], subset_ratio, seed)
        self.train_d = self.train_d[:, ix]
        self.valid_d = self.valid_d[:, ix]
        self.test_d = self.test_d[:, ix]

##############################################################################################################


class BalancedData(Data):
    def __init__(self):
        super(BalancedData, self).__init__()
        self._splitted = False

    def getbatch(self, batch_size, source='train', balanced=False):
        if balanced:
            return self.getbatch_balanced(batch_size)  # only train
        else:
            return super(BalancedData, self).getbatch(batch_size, source=source)

    def _init_balanced(self):
        self._b_mask = [(None)] * self.n_lab
        self._b_count = np.empty((self.n_lab), dtype=np.int)
        self._b_pos = np.zeros((self.n_lab), dtype=np.int)
        self._b_perm = [(None)] * self.n_lab
        for k in xrange(self.n_lab):
            mask = np.flatnonzero(self.train_l == k)
            self._b_mask[k] = mask
            self._b_count[k] = len(mask)
            self._b_perm[k] = self.rng.permutation(self._b_count[k])

        self._splitted = True

    def _get_save_slice(self, perm, pos, batch_size):
        if (pos + batch_size) < len(perm):
            pos = pos + batch_size
            return perm[pos - batch_size:pos], pos
        else:
            slice = perm[pos:]
            pos = batch_size - len(slice)
            slice = np.hstack((slice, perm[:pos]))
            return slice, 0  # need to shuffle

    def getbatch_balanced(self, batch_size):
        if not self._splitted:
            self._init_balanced()
        data = np.empty((batch_size, self.train_d.shape[1]), dtype=np.float32)
        label = np.empty((batch_size), dtype=np.int16)
        batch_size = batch_size // self.n_lab

        for k, (mask, perm, pos) in enumerate(zip(self._b_mask, self._b_perm, self._b_pos)):
            slice, pos = self._get_save_slice(perm, pos, batch_size)
            d = self.train_d[mask[slice]]
            l = self.train_l[mask[slice]]
            self._b_pos[k] = pos
            if pos == 0:
                self._b_perm[k] = self.rng.permutation(self._b_count[k])
                self._b_pos[k] = 0

            data[k * batch_size:(k + 1) * batch_size] = d
            label[k * batch_size:(k + 1) * batch_size] = l

        return data, label

        ##############################################################################################################


class QueueData(Data):
    def __init__(self):
        super(QueueData, self).__init__()
        ### Queue stuff ###
        self._queue_prio = np.zeros(self._training_count)
        self._queue_last = np.zeros(self._training_count)
        self._queue_ix = np.arange(self._training_count)
        self._queue_count = np.ones(self._training_count)  # Initialise to one s.t. 1/count is possible
        self._batch_ix = None

    def queueget(self, n):
        assert self._batch_ix is None, "Update Priorities first before requesting new batch"
        slice = self._queue_ix[:n]  # indices of n highest elements
        self._batch_ix = slice  # store for updates
        if isinstance(self.train_d, np.ndarray):
            ret = (self.train_d[slice], self.train_l[slice], self._queue_count[slice])
        elif isinstance(self.train_d, list):
            data = [self.train_d[i] for i in slice]
            label = [self.train_l[i] for i in slice]
            count = [self._queue_count[i] for i in slice]
            ret = (data, label, count)

        self._queue_count[slice] += 1
        return ret

    def queueupdate(self, nlls, iteration):
        assert self._batch_ix.shape == nlls.shape, "Cannot update, indices not known"
        self._queue_prio[self._batch_ix] = nlls/(1+0.017*self._queue_count[self._batch_ix]) \
                                                -0.004*(iteration - self._queue_last[self._batch_ix])
        #    for i,nll in zip(self._batch_ix, nlls): # Update priorities in original list
        #      p = nll/(1+0.017*self._queue_count[i]) - 0.004*(iteration - self._queue_last[i])
        #      self._queue_prio[i] = p

        self._queue_ix = np.argsort(self._queue_prio)[::-1]  # restore order, high prios first
        self._batch_ix = None

    def queuereset(self):
        self._queue_prio = np.zeros(self._training_count)
        self._queue_ix = np.arange(self._training_count)
        self._queue_count = np.ones(self._training_count)  # Initialise to one s.t. 1/count is possible
        self._batch_ix = None

    def _getdata(self):
        return (self.train_d, self.train_l, self.valid_d, self.valid_l)

#  def swapSets(self):
#    train_d, train_l, valid_d, valid_l = self._getdata()
#    self.train_d, self.train_l, self.valid_d, self.valid_l = valid_d, valid_l, train_d, train_l,
#    self._training_count = self.train_d.shape[0]
#    self._perm = np.arange(self._training_count)
#    self._queue_prio = np.zeros(self._training_count)
#    self._queue_last = np.zeros(self._training_count)
#    self._queue_ix   = np.arange(self._training_count)
#    self._queue_count= np.ones(self._training_count) # Initialise to one s.t. 1/count is possible

    def save(self, path="data"):
        f = file(path, 'w')
        if isinstance(self.train_d, np.ndarray):
            cPickle.dump(self._getdata(), f, protocol=2)
        elif isinstance(self.train_d, list):
            dat = self.valid_d + self.train_d
            lab = self.valid_l + self.train_l
            for (d, l) in zip(dat, lab):
                cPickle.dump(d, f, protocol=2)
                cPickle.dump(l, f, protocol=2)
        f.close()


### Toy Data Sets ############################################################################################
class AdultData(Data):
    def __init__(self, path='~/devel/data/adult.pkl', create=False):
        path = os.path.expanduser(path)
        if create:
            self._fields = 'age,workclass,fnlwgt,education,educationnum,maritalstatus,occupation,relationship,race,sex,capitalgain,capitalloss,hoursperweek,nativecountry,target'.split(',')
            self._kinds = 'cont,cat,cat,cat,cont,cat,cat,cat,cat,cat,cont,cont,cont,cat,cat'.split(',')

            data_socket = urllib2.urlopen('http://archive.ics.uci.edu/ml/machine-learning-databases/adult/adult.data')
            train_d = np.genfromtxt(data_socket, skip_header=1, delimiter=',', names=self._fields, dtype=None)
            train_d = self._normalise_adult(train_d)
            self.train_l = train_d[:, -1].astype('int16')  # np.expand_dims(train_d[:,-1].astype('int16'), 1)
            self.train_d = train_d[:, :-1]

            test_socket = urllib2.urlopen('http://archive.ics.uci.edu/ml/machine-learning-databases/adult/adult.test')
            valid_d = np.genfromtxt(test_socket, skip_header=1, delimiter=',', names=self._fields, dtype=None)
            valid_d = self._normalise_adult(valid_d)
            self.valid_l = valid_d[:, -1].astype('int16')  #np.expand_dims(valid_d[:,-1].astype('int16'), 1)
            self.valid_d = valid_d[:, :-1]

            ut.pickleSave((self.train_d, self.train_l, self.valid_d, self.valid_l), path)

        else:
            self.train_d, self.train_l, self.valid_d, self.valid_l = ut.pickleLoad(path)

        super(AdultData, self).__init__()

    def _normalise_adult(self, data):
        ret = np.zeros((data.size, len(self._fields)), dtype='float32')
        for i, (name, kind) in enumerate(zip(self._fields, self._kinds)):
            if kind == 'cat':
                unique = np.unique(data[name])
                for k, val in enumerate(unique):
                    ret[(data[name] == val), i] = k
                ret[:, i] *= 1 / ret[:, i].max()
            elif kind == 'cont':
                ret[:, i] = data[name].astype('float32')
                ret[:, i] *= 1 / ret[:, i].max()
            else:
                print name + ' has no kind specified'
        return ret

##########################################################################################


class MNISTData(Data):
    def __init__(self, path=None, convert2image=True, warp_on=False, shift_augment=True, center=True):
        if path is None:
            (self.train_d, self.train_l), (self.valid_d, self.valid_l), (self.test_d, self.test_l) = self.download()
        else:
            path = os.path.expanduser(path)
            (self.train_d, self.train_l), (self.valid_d, self.valid_l), (self.test_d, self.test_l) = ut.pickleLoad(path)

        self.warp_on = warp_on
        self.shif_augment = shift_augment
        self.return_flat = not convert2image
        self.test_l = self.test_l.astype(np.int16)
        self.train_l = self.train_l.astype(np.int16)
        self.valid_l = self.valid_l.astype(np.int16)

        if center:
            self.test_d -= self.test_d.mean()
            self.train_d -= self.train_d.mean()
            self.valid_d -= self.valid_d.mean()

        self.convert_to_image()
        if self.shif_augment:
            self._stripborder(1)
            self.train_d, self.train_l = self._augmentMNIST(self.train_d, self.train_l, crop=2, factor=4)

        super(MNISTData, self).__init__()
        if not convert2image:
            self.example_shape = self.train_d[0].size

        print "MNIST data is converted/augmented to shape", self.example_shape


    def download(self):
        if os.name == 'nt':
            dest = os.path.join(os.environ['APPDATA'], 'ELEKTRONN')
        else:
            dest = os.path.join(os.path.expanduser('~'), '.ELEKTRONN')

        if not os.path.exists(dest):
            os.makedirs(dest)

        dest = os.path.join(dest, 'mnist.pkl.gz')

        if os.path.exists(dest):
            print "Found existing mnist data"
            return ut.pickleLoad(dest)
        else:
            print "Downloading mnist data from"
            print "http://www.elektronn.org/downloads/mnist.pkl.gz"
            f = urllib2.urlopen("http://www.elektronn.org/downloads/mnist.pkl.gz")
            data = f.read()
            print "Saving data to %s" %(dest,)
            with open(dest, "wb") as code:
                code.write(data)

            return ut.pickleLoad(dest)



    def convert_to_image(self):
        """For MNIST / flattened 2d, single-layer, square images"""

        valid_size = self.valid_l.size
        test_size = self.test_l.size
        data = np.vstack((self.valid_d, self.test_d, self.train_d))
        size = data[0].size
        n = int(np.sqrt(size))
        assert abs(n**2 - size) < 1e-6, '<convertToImage> data is not square'
        count = data.shape[0]
        data = data.reshape((count, 1, n, n))
        self.valid_d = data[:valid_size]
        self.test_d = data[valid_size:valid_size + test_size]
        self.train_d = data[valid_size + test_size:]

    def getbatch(self, batch_size, source='train'):
        if source == 'valid':
            ret = super(MNISTData, self).getbatch(batch_size, 'valid')
        if source == 'test':
            ret = super(MNISTData, self).getbatch(batch_size, 'test')
        else:
            d, l = super(MNISTData, self).getbatch(batch_size, source)
            if self.warp_on:
                d = self._warpaugment(d)
            ret = d, l

        if self.return_flat:
            ret = (ret[0].reshape((batch_size, -1)), ret[1])

        return ret

    def _stripborder(self, pix=1):
        s = self.train_d.shape[-1]
        self.valid_d = self.valid_d[:, :, pix:s - pix, pix:s - pix]
        self.test_d = self.test_d[:, :, pix:s - pix, pix:s - pix]

    def _warpaugment(self, d, amount=1):
        rot_max = 5 * amount
        shear_max = 7 * amount
        scale_max = 1.15 * amount
        stretch_max = 0.25 * amount

        shear = shear_max * 2 * (np.random.rand() - 0.5)
        twist = rot_max * 2 * (np.random.rand() - 0.5)
        rot = 0  # min(rot_max - abs(twist), rot_max  * (np.random.rand()))
        scale = 1 + (scale_max - 1) * np.random.rand(2)
        stretch = stretch_max * 2 * (np.random.rand(4) - 0.5)

        ps = (d.shape[0], ) + d.shape[2:]
        w = warping.warp3dFast(d, ps, rot, shear, (scale[0], scale[1], 1), stretch, twist)
        return w

    def _augmentMNIST(self, data, label, crop=2, factor=4):
        """
        Creates new data, by cropping/shifting data.
        Control blow-up by factor and maximum offset by crop
        """
        n = data.shape[-1]
        new_size = (n - crop)
        new_data = np.zeros((0, 1, new_size, new_size), dtype=np.float32)  # store new data in here
        new_label = np.zeros((0, ), dtype=np.int16)
        pos = [(i % crop, int(i / crop) % crop) for i in xrange(crop**2)]  # offests of different positions
        perm = np.random.permutation(xrange(crop**2))

        for i in xrange(factor):  # create <factor> new version of data
            ix = pos[perm[i]]
            new = (data[:, :, ix[0]:ix[0] + new_size, ix[1]:ix[1] + new_size])
            new_data = np.concatenate((new_data, new), axis=0)
            new_label = np.concatenate((new_label, label), axis=0)

        return new_data, new_label


class BuzzData(Data):
    def __init__(self, path='~/devel/data/Buzz/Twitter/twitter.pkl', norm_targets=True, target_scale=9999, fold_no=0):
        path = os.path.expanduser(path)
        data, target = ut.pickleLoad(path)
        #    N = len(data)
        #    data = data.reshape((N, -1))
        #    data = data[:8000]
        #    target = target[:8000]
        if norm_targets:
            target /= target.max()
        if target_scale is not None:
            target = np.log10(target * target_scale + 1)

        super(BuzzData, self).createCVSplit(data, target, use_fold=fold_no)
        super(BuzzData, self).__init__()
        self.example_shape = data.shape[-1]
        self.n_taps = data.shape[-2]
        self.n_lab = 1

    def getbatch(self, batch_size, source='train'):
        d, l = super(BuzzData, self).getbatch(batch_size, source=source)
        l = l[:, None]
        return d, l


class PianoData(Data):
    def __init__(self, path='~/devel/data/PianoRoll/Nottingham_enc.pkl', n_tap=20, n_lab=58):
        path = os.path.expanduser(path)
        (self.train_d, self.valid_d, self.test_d) = ut.pickleLoad(path)
        super(PianoData, self).__init__(n_lab=n_lab)
        self.example_shape = self.train_d[0].shape[-1]
        self.n_taps = n_tap
        self.n_lab = n_lab

    def getbatch(self, batch_size, source='train'):
        if source == 'train':
            if (self._pos + batch_size) < self._training_count:
                self._pos += batch_size
                slice = self._perm[self._pos - batch_size:self._pos]
            else:  # get new permutation
                self._perm = self.rng.permutation(self._training_count)
                self._pos = 0
                slice = self._perm[:batch_size]

            data = [self.train_d[i] for i in slice]

        elif source == 'valid':
            data = self.valid_d[:batch_size]

        elif source == 'test':
            data = self.test_d[:batch_size]

        lengths = np.array(map(len, data))
        start_t = np.round(np.random.rand(batch_size) * (lengths - self.n_taps - 1)).astype(np.int)
        x = np.array([d[t:t + self.n_taps].astype(np.float32) for d, t in zip(data, start_t)])
        y = np.array([d[t + self.n_taps] for d, t in zip(data, start_t)])
        return x, y


class GeneData(Data):
    def __init__(self, path='~/devel/data/GEMLeR_GeneExpression/Breast_Colon.pkl', fold_no=0):
        path = os.path.expanduser(path)
        data, target = ut.pickleLoad(path)
        super(GeneData, self).createCVSplit(data, target, use_fold=fold_no)
        super(GeneData, self).__init__()
        self.example_shape = data.shape[-1]
        self.n_lab = 1

    def getbatch(self, batch_size, source='train'):
        d, l = super(GeneData, self).getbatch(batch_size, source=source)
        l = l[:, None]
        return d, l


if __name__ == "__main__":
    #  from matplotlib import pyplot as plt
    #  from Net.introspection import embedMatricesInGray   
    #  data = MNISTData( warp_on=True)  
    #  d, l = data.getbatch(200, 'train')
    #  m = embedMatricesInGray(d[:,0])
    #  plt.imshow(m, interpolation='none', cmap='gray')
    data = MNISTData(path=None, convert2image=False, shift_augment=False)

    #  data = PianoData(n_tap=20, n_lab=58)
    d, l = data.getbatch(10)
    data = AdultData()



# -*- coding: utf-8 -*-
"""
ELEKTRONN - Neural Network Toolkit

Copyright (c) 2014 - now
Max-Planck-Institute for Medical Research, Heidelberg, Germany
Authors: Marius Killinger, Gregor Urban
"""
# CONFIGURATION MNIST_example_warp

### Pipeline Setup ### ------------------------------------------------------------------------------------------------------------
save_path = "~/CNN_Training/2D/"  # (*) <String>: where to create the CNN directory.
# In this directory a new folder is created with the save name of the model

### Paths and General ### ------------------------------------------------------------------------------------------------------------
save_name = 'MNIST_example_warp'  # (*) <String>: with the save_name.
save_path += save_name + '/'
param_file = None  # <String>/<None>: optional path to parameter file to initialise weights from

### Network Architecture ### ------------------------------------------------------------------------------------------------------------
# Note: the last layer is added automatically (with n_lab outputs)
activation_func = 'relu'  # <String> or <List> of <String>s: [relu], abs, linear, sig, tanh, (globally or per layer)
batch_size = 50  # (*) <Int>: number of different slices/examples used for one optimisation step
dropout_rates = None  # (*) <List> of <Float>>(0,1) or <Float>(0,1): "fail"-rates (globally or per layer)
# The last layer never has dropout.
# Other
MLP_layers = [
    300, 300
]  # (*) <List> of <Int>: numbers of filters for perceptron layers (after conv layers)
target = 'nll'  # <String>: 'nll' or 'regression'

### Data General ### ---------------------------------------------------------------------------------------------------------------------
mode = 'vect-scalar'  # (*) <String>: combination of data and label types: 'img-img', 'img-scalar', vect-scalar'
n_lab = None  # (*) <Int>/<None>: (None means auto detect, very slow, don't do this!)
background_processes = 3  # <Bool>/<Int>: whether to "pre-fetch" batches in separate background
# process, <Bool> or number of processes (True-->2)

### Data Alternative / vect-scalar ### (this may replace the above CNN block) ### -------------------------------------------------------
data_class_name = 'MNISTData'  # Name of Data Class in traindata module
data_load_kwargs = dict(
    path=None,
    convert2image=False,
    warp_on=True,
    shift_augment=True)
data_batch_kwargs = dict()  # <Dict>: Arguments for getbach method of Data Class (for training set only!)
# The batch_size argument is added internally and needn't be specified here

### Optimisation Options ### ------------------------------------------------------------------------------------------------------------
n_steps = 100000  # (*) <Int>: number of update steps
max_runtime = 30 * 60  # (*) <Int>: maximal Training time in seconds (overrides n_steps)
history_freq = [
    100
]  # (*) <List> of single <Int>: create plots, print status and test model after x steps
monitor_batch_size = 10000  # (*) <Int>: number of patches to test model on (valid and train subset)

weight_decay = False  # ($) False/<Float>: weighting of L2-norm on weights ("lambda"), False is equal to 0.0

optimizer = 'SGD'  # ($) <String>: [SGD]/CG/RPORP/LBFGS method for training

LR_decay = 0.99  # (*) <Float>: decay of SGD learning rate w.r.t to an interval of 1000 update steps
LR_schedule = None  # (*) <List> of tuples/<None>: (#iteration, new_LR), if used this sets the LR
# at the specified iteration steps to the specified value. This is independent of the decay.

# ---------------------------------------------------------------------------------------------------------------------------------------
### SGD
SGD_params = dict(LR=0.02,
                  momentum=0.9
                  )  # (*) <Dict>: initial learning rate and momentum



# -*- coding: utf-8 -*-
# ELEKTRONN - Neural Network Toolkit
#
# Copyright (c) 2014 - now
# Max-Planck-Institute for Medical Research, Heidelberg, Germany
# Authors: Marius Killinger, Gregor Urban
"""
Supplementary functions to initialise CNN-params with gabor filters
"""

import numpy as np
from matplotlib import pyplot as plt

#def makeGabor(filter_angle, n_modes, size, offset):
#    """
#    filter_angle: in degree: 0 to 180
#    n_modes = 1,2,3 etc.
#    size: filter size
#    offset: 0 to 180
#    """
#    sigma = 0.35 # 0.5
#    freq  = n_modes / np.pi  *1.6 #n_modes / np.pi * 1.3 # 1.5
#    assert filter_angle<360 and filter_angle>=0
#    xMax = yMax = (size-1)/2+1 if size%2==1 else size/2+1
#    xMin = yMin = -xMax+1
#    assert yMax>0
#    assert xMax>0
#    sigmaX = sigma
#    sigmaY = sigma
#    theta = float(filter_angle)/180.0*np.pi
#    m_offset = float(offset)/180.0*np.pi
#    gabor  = np.empty((int(xMax)-int(xMin), int(yMax)-int(yMin)), dtype=np.float32)
#    #print shape(gabor)
#    for x in range(int(xMin), int(xMax)):
#        for y in range(int(yMin), int(yMax)):
#            xPrime =  float(x)/xMax*np.cos(theta)+float(y)/yMax*np.sin(theta)
#            yPrime = -float(x)/xMax*np.sin(theta)+float(y)/yMax*np.cos(theta)
#            gabor[x-int(xMin),y-int(yMin)] = np.exp(-0.5*((xPrime*xPrime)/(sigmaX*sigmaX)+(yPrime*yPrime)/(sigmaY*sigmaY)))*np.cos(2.0*np.pi*(freq*xPrime + m_offset))
#
#    gabor = gabor - gabor.mean()
#    gabor = gabor / np.square(gabor).sum()
#    return gabor


def makeGabor(filter_angle, n_modes, size, offset):
    """
    Parameters
    ----------
    
    filter_angle: in degree: 0 to 180
    n_modes = 1,2,3 etc.
    size: filter size
    offset: 0 to 180
    """
    assert 360 > filter_angle >= 0
    sigma = 0.35  # 0.5
    freq = n_modes / np.pi * 1.6  #n_modes / np.pi * 1.3 # 1.5
    theta = float(filter_angle) / 180.0 * np.pi
    m_offset = float(offset) / 180.0 * np.pi
    Max = (size - 1) / 2 if size % 2 == 1 else size / 2

    X, Y = np.meshgrid(np.linspace(-Max, Max, size), np.linspace(-Max, Max, size))
    X_ = X / (Max + 1) * np.cos(theta) + Y / (Max + 1) * np.sin(theta)
    Y_ = -X / (Max + 1) * np.sin(theta) + Y / (Max + 1) * np.cos(theta)
    gabor = np.exp(-0.5 * ((X_**2) + (Y_**2)) / (sigma**2)) * np.cos(2.0 * np.pi * (freq * Y_ + m_offset))
    gabor = gabor - gabor.mean()
    gabor = gabor / np.square(gabor).sum()
    return gabor.astype(np.float32)


def makeGaborFilters(size, number):
    """
    Use this to generate ``number`` first order
    and ``number`` second order filters
    """
    nd2 = number  #int(float(number)*0.1)
    nd3 = number  #number-nd2
    ret = []
    for theta in np.arange(0, 360, (360.0 / nd2)):
        ret.append(makeGabor(theta, 1, size, 45))
        #ret.append( makeGabor(theta, 1 ,size, np.random.random()*360) )
    for theta in np.arange(0, 180, (180.0 / nd3)):
        ret.append(makeGabor(theta, 2, size, 0))
        #ret.append( makeGabor(theta, 2 ,size, 45 + (np.random.random()-0.5)*8) )
    return np.array(ret)


def blob(size):
    """
  Return Gaussian blob filter
  """
    x, y = np.meshgrid(np.linspace(-2, 2, size), np.linspace(-2, 2, size))
    ret = np.exp(-0.5 * (np.square(x) + np.square(y)))
    ret = ret / np.square(ret).sum()
    return ret


if __name__ == '__main__':
    from introspection import embedMatricesInGray
    ga = makeGaborFilters(3, 12)
    #ga = np.maximum(0, ga)
    mat = embedMatricesInGray(ga, 1)
    plt.imshow(mat, interpolation='none')
    plt.gray()

    ga = makeGaborFilters(4, 12)
    mat2 = embedMatricesInGray(ga, 1)
    plt.figure()
    plt.imshow(mat2, interpolation='none')
    plt.gray()



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

import os
import setuptools
from setuptools import setup, find_packages, Extension
from pkg_resources import parse_version

# Setuptools >=18.0 is needed for Cython to work correctly.
if parse_version(setuptools.__version__) < parse_version('18.0'):
    raise ImportError(
        'Your installed Setuptools version is too old.\n'
        '  Please upgrade it to at least 18.0, e.g. by running\n'
        '  "pip2 install --upgrade setuptools"\n'
        '  or if that fails:\n'
        '  "pip install --upgrade setuptools".\n'
        '  If both of them fail, try additionally passing the "--user" switch to the install commands, or use Anaconda2.')


def read(fname):
    """
    Utility function to read the README file.
    Used for the long_description.  It's nice, because now 1) we have a top level
    README file and 2) it's easier to type in the README file than to put a raw
    string in below ...
    """
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


ext_modules = [
    Extension('elektronn.training._warping',
              sources=['elektronn/training/_warping.pyx'],
              extra_compile_args=['-std=c99', '-O3', '-Wall', '-Wextra'])
]

setup(
    name='elektronn',
    version='1.0.7',
    packages=find_packages(),
    scripts=['elektronn/scripts/elektronn-train',
             'elektronn/scripts/elektronn-profile',
    ],
    ext_modules=ext_modules,
    setup_requires=['cython>=0.23'],
    install_requires=[
        'numpy>=1.8',
        'scipy>=0.14',
        'matplotlib>=1.4',
        'h5py>=2.2',
        'theano>=0.7',
        'cython>=0.23',
    ],
    extras_require={'cross-validation': ['scikit-learn']},
    author="Marius Killinger, Gregor Urban",
    author_email="Marius.Killinger@mailbox.org",
    description=("A highly configurable toolkit for training 3d/2d CNNs and general Neural Networks"),
    long_description=read('README.rst'),
    license="GPLv3",
    keywords="cnn theano convolutional neural network machine learning classification",
    url="http://www.elektronn.org/",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2 :: Only',
        'Programming Language :: Python :: 2.7',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Information Analysis',
    ],
)



# -*- coding: utf-8 -*-
# ELEKTRONN - Neural Network Toolkit
#
# Copyright (c) 2014 - now
# Max-Planck-Institute for Medical Research, Heidelberg, Germany
# Authors: Marius Killinger, Gregor Urban

import time
import numpy as np
import theano
import theano.tensor as T
from theano.tensor.nnet.conv3d2d import conv3d

import pooling
from netutils import initWeights
try:
    from malis.malisop import malis_weights
except:
    malis_weights = None


def getOutputShape(insh, fsh, pool, mfp, r=1):
    """
    Returns shape of convolution result from (bs, z, ch, x, y) * (nof, z, ch, xf, yf)
    """
    if insh[0] is None:
        bs = None
    else:
        bs = insh[0] * np.prod(pool) if mfp else insh[0]
    z = (insh[1] - fsh[1] + 1) // pool[0]
    ch = fsh[0] if r == 1 else fsh[0] // r
    x = (insh[3] - fsh[3] + 1) // pool[1]
    y = (insh[4] - fsh[4] + 1) // pool[2]
    return (bs, z, ch, x, y)


def getProbShape(output_shape, mfp_strides):
    """
  Given outputshape (bs, z, ch, x, y) and mfp_stride (sx, sy) returns shape of Class Prob output
  """
    if mfp_strides is None:
        return (output_shape[0], output_shape[2], output_shape[1], output_shape[3], output_shape[4])
    else:
        return (1, output_shape[2], output_shape[1] * mfp_strides[0],
                output_shape[3] * mfp_strides[1], output_shape[4] * mfp_strides[2])


class ConvLayer3d(object):
    """
    Conv-Pool Layer of a CNN

    :type input: theano.tensor.dtensor5 ('batch', z, 'channel', x, y)
    :param input: symbolic image tensor, of shape input_shape

    :type input_shape: tuple or list of length 5
    :param input_shape: (batch size, z, num input feature maps,  y, x)

    :type filter_shape: tuple or list of length 5
    :param filter_shape: (number of filters, filter z, num input feature maps, filter y,filter x)


    :type pool: int 3-tuple
    :param pool: the down-sampling (max-pooling) factor

    :type activation_func: string
    :param activation_func: Options: tanh, relu, sig, abs, linear, maxout <i>

    :type enable_dropout: Bool
    :param enable_dropout: whether to enable dropout in this layer. The default rate is 0.5 but it can be changed with
               self.activation_noise.set_value(set_value(np.float32(p)) or using cnn.setDropoutRates

    :type use_fragment_pooling: Bool
    :param use_fragment_pooling: whether to use max fragment pooling in this layer (MFP)

    :type reshape_output: Bool
    :param reshape_output: whether to reshape class_probabilities to (bs, cls, x, y) and re-assemble fragments
                         to dense images if MFP was enabled. Use this in for the last layer.

    :type mfp_offsets: list of list of ints
    :param mfp_offsets: this lists specifies the offsets that the MFP-fragments have w.r.t to the original patch.
                      Only needed if MFP is enabled.

    :type mfp_strides: list of int
    :param mfp_strides: the strides of the output in each dimension

    :type input_layer: layer object
    :param input_layer: just for keeping track of un-usual input layers

    :type W: np.ndarray or T.TensorVariable
    :param W: weight matrix. If array, the values are used to initialise a shared variable for this layer.
                           If TensorVariable, than this variable is directly used (weight sharing with the
                           layer from which this variable comes from)

    :type b: np.ndarray or T.TensorVariable
    :param b: bias vector. If array, the values are used to initialise a shared variable for this layer.
                           If TensorVariable, than this variable is directly used (weight sharing with the
                           layer from which this variable comes from)

    :type pooling_mode: str
    :param pooling_mode: 'max' or 'maxabs' where the first is normal maxpooling and the second also retains
                       sign of large negative values

    """

    def __init__(self,
                 input,
                 input_shape,
                 filter_shape,
                 pool,
                 activation_func,
                 enable_dropout,
                 use_fragment_pooling,
                 reshape_output,
                 mfp_offsets,
                 mfp_strides,
                 input_layer=None,
                 W=None,
                 b=None,
                 pooling_mode='max',
                 affinity=False):

        assert len(filter_shape) == 5
        assert input_shape[2] == filter_shape[2]

        self.input = input
        self.pool = pool
        self.number_of_filters = filter_shape[0]
        self.filter_shape = filter_shape
        self.activation_func = activation_func
        self.input_shape = input_shape
        self.input_layer = input_layer
        self.mfp_strides = mfp_strides
        self.mfp_offsets = mfp_offsets
        self.reshape_output = reshape_output

        print "3DConv: input=", input_shape, "\tfilter=", filter_shape  #,"@std=",W_bound

        if W is None:
            W_values = np.asarray(
                initWeights(filter_shape,
                            scale='glorot',
                            mode='normal',
                            pool=pool),
                dtype=theano.config.floatX)
            self.W = theano.shared(W_values, name='W_conv', borrow=True)
        else:
            if isinstance(W, np.ndarray):
                self.W = theano.shared(
                    W.astype(np.float32),
                    name='W_conv',
                    borrow=True)
            else:
                assert isinstance(
                    W,
                    T.TensorVariable), "W must be either np.ndarray or theano var"
                self.W = W

        # the bias is a 1D tensor -- one bias per output feature map
        if activation_func in ['ReLU', 'relu']:
            norm = filter_shape[1] * filter_shape[3] * filter_shape[4]  #
            b_values = np.ones(
                (filter_shape[0], ),
                dtype=theano.config.floatX) / norm
        if b is None:
            n_out = filter_shape[0]
            if activation_func == 'relu' or activation_func == 'ReLU':
                norm = filter_shape[1] * filter_shape[3] * filter_shape[4]
                b_values = np.asarray(
                    initWeights(
                        (n_out, ),
                        scale=1.0 / norm,
                        mode='const'),
                    dtype=theano.config.floatX)
            elif activation_func == 'sigmoid':
                b_values = np.asarray(
                    initWeights(
                        (n_out, ),
                        scale=0.5,
                        mode='const'),
                    dtype=theano.config.floatX)
            else:  # activation_func=='tanh':
                b_values = np.asarray(
                    initWeights(
                        (n_out, ),
                        scale=1e-6,
                        mode='fix-uni'),
                    dtype=theano.config.floatX)

            self.b = theano.shared(value=b_values, borrow=True, name='b_conv')

        else:
            if isinstance(b, np.ndarray):
                self.b = theano.shared(
                    b.astype(np.float32),
                    name='b_conv',
                    borrow=True)
            else:
                assert isinstance(
                    b,
                    T.TensorVariable), "b must be either np.ndarray or theano var"
                self.b = b

        # store parameters of this layer
        self.params = [self.W, self.b]

        # convolve input feature maps with filters
        self.mode = theano.compile.get_default_mode()
        self.conv_out = conv3d(
            signals=input,
            filters=self.W,
            border_mode='valid',
            filters_shape=filter_shape
        )  # signals_shape=input_shape if input_shape[0] is not None else None)

        # down-sample each feature map individually, using maxpooling
        if np.any(pool != 1):
            pool_func = lambda x: pooling.pooling3d(x, pool_shape=pool, mode=pooling_mode)
            if use_fragment_pooling:
                pooled_out, self.mfp_offsets, self.mfp_strides = self.fragmentpool(
                    self.conv_out, pool, mfp_offsets, mfp_strides, pool_func)
            else:
                pooled_out = pool_func(self.conv_out)
        else:
            pooled_out = self.conv_out

        if enable_dropout:
            print "Dropout: ACTIVE"
            self.activation_noise = theano.shared(
                np.float32(0.5),
                name='Dropout Rate')
            rng = T.shared_randomstreams.RandomStreams(int(time.time()))
            p = 1 - self.activation_noise
            self.dropout_gate = 1.0 / p * rng.binomial(
                (pooled_out.shape[1], pooled_out.shape[3],
                 pooled_out.shape[4]),
                1,
                p,
                dtype=theano.config.floatX)
            pooled_out = pooled_out * self.dropout_gate.dimshuffle(('x', 0, 'x', 1, 2))

        lin_output = pooled_out + self.b.dimshuffle('x', 'x', 0, 'x', 'x')
        self.lin_output = lin_output
        r = 1
        if activation_func == 'tanh':
            self.activation_func = 'tanh'
            self.output = T.tanh(lin_output)  # shape: (batch_size, num_outputs)
        elif activation_func in ['ReLU', 'relu']:  #rectified linear unit
            self.activation_func = 'relu'
            self.output = lin_output * (lin_output > 0)  # shape: (batch_size, num_outputs)
        elif activation_func in ['linear', 'none', 'None', None]:
            self.activation_func = 'linear'
            self.output = lin_output
        elif activation_func in ['abs']:
            self.activation_func = 'abs'
            self.output = T.abs_(lin_output)
        elif activation_func in ['sigmoid']:
            self.activation_func = 'sigmoid'
            self.output = T.nnet.sigmoid(lin_output)
        elif activation_func.startswith("maxout"):
            r = int(activation_func.split(" ")[1])
            assert r >= 2
            self.output = pooling.maxout(lin_output, factor=r, axis=2)
        else:
            raise NotImplementedError()

        output_shape = getOutputShape(
            (1 if input_shape[0] is None else input_shape[0], ) +
            input_shape[1:], filter_shape, pool, use_fragment_pooling, r)

        print "Output=", output_shape, "Dropout", (
            "ON," if enable_dropout else
            "OFF,"), "Act:", activation_func, "pool:", pooling_mode
        self.output_shape = output_shape  # e.g. (None, 16, 100, 100)

        if affinity:
            raise RuntimeError("Dont use this code")
#      self.class_probabilities = T.nnet.sigmoid(lin_output) # (bs, z, 3, x, y)
#      self.class_probabilities = self.class_probabilities.dimshuffle((0,2,1,3,4))
#      sh = lin_output.shape
#      if use_fragment_pooling:     
#        self.fragmentstodense(sh) # works on
#
#      self.prob_shape = getProbShape(output_shape, self.mfp_strides)
#      self.class_prediction = T.gt(self.class_probabilities, 0.5)
#      # self.class_probabilities = (bs,3,z,x,y)

        else:
            sh = lin_output.shape  #(bs,x,ch,y,z)  # use this shape to reshape the output to image-shape after softmax
            sh = (sh[2], sh[0], sh[1], sh[3], sh[4])  #(ch, x, y, bs)
            #  put spatial, at back --> (ch,bs,x,y,z), flatten this --> (ch, bs*x*y*z), swap labels --> (bs*x*y*z, ch)
            self.class_probabilities = T.nnet.softmax(
                    lin_output.dimshuffle((2, 0, 1, 3, 4)).flatten(2).dimshuffle((1, 0)))
            if reshape_output:
                self.reshapeoutput(sh)
                if use_fragment_pooling:
                    self.fragmentstodense(sh)

                self.prob_shape = getProbShape(output_shape, self.mfp_strides)
                print "Class Prob Output =", self.prob_shape
            # compute prediction as class whose "probability" is maximal in symbolic form
            self.class_prediction = T.argmax(self.class_probabilities, axis=1)

    def fragmentpool(self, conv_out, pool, offsets, strides, pool_func):
        result = []
        offsets_new = []
        if offsets is None:
            offsets = np.array([[0, 0, 0]])
        if strides is None:
            strides = [1, 1, 1]

        sh = conv_out.shape

        for pz in xrange(pool[0]):
            for px in xrange(pool[1]):
                for py in xrange(pool[2]):
                    pzxy = pool_func(conv_out[:, pz:sh[1] - pool[
                        0] + pz + 1, :, px:sh[3] - pool[1] + px + 1, py:sh[4] -
                                              pool[2] + py + 1])

                    result.append(pzxy)
                    for p in offsets:
                        new = p.copy()
                        new[0] += pz * strides[0]
                        new[1] += px * strides[1]
                        new[2] += py * strides[2]
                        offsets_new.append(new)

        result = T.concatenate(result, axis=0)
        offsets_new = np.array(offsets_new)
        strides = np.multiply(pool, strides)

        return result, offsets_new, strides

    def fragmentstodense(self, sh):
        mfp_strides = self.mfp_strides
        example_stride = np.prod(mfp_strides)  # This stride is conceptually unneeded but theano-grad fails otherwise
        zero = np.array((0), dtype=theano.config.floatX)
        embedding = T.alloc(zero, 1, sh[0], sh[2] * mfp_strides[0], sh[3] *
                            mfp_strides[1], sh[4] * mfp_strides[2])
        ix = self.mfp_offsets
        for i, (n, m, k) in enumerate(ix):
            embedding = T.set_subtensor(
                embedding[:, :, n::mfp_strides[0], m::mfp_strides[1], k::mfp_strides[2]],
                self.class_probabilities[i::example_stride])
        self.class_probabilities = embedding

    def reshapeoutput(self, sh):
        # same shape as result, no significant time cost
        self.class_probabilities = self.class_probabilities.dimshuffle((1, 0)).reshape(sh).dimshuffle((1, 2, 0, 3, 4))
        self.class_probabilities = self.class_probabilities.dimshuffle((0, 2, 1, 3, 4))

    def randomizeWeights(self, scale='glorot', mode='uni'):
        n_out = self.filter_shape[0]
        if self.activation_func == 'relu':
            norm = self.filter_shape[1] * self.filter_shape[3] * self.filter_shape[4]
            b_values = np.asarray(
                initWeights(
                    (n_out, ),
                    scale=1.0 / norm,
                    mode='const'),
                dtype=theano.config.floatX)
        elif self.activation_func == 'sigmoid':
            b_values = np.asarray(
                initWeights(
                    (n_out, ),
                    scale=0.5,
                    mode='const'),
                dtype=theano.config.floatX)
        else:  #self.activation_func=='tanh':
            b_values = np.asarray(
                initWeights(
                    (n_out, ),
                    scale=1e-6,
                    mode='fix-uni'),
                dtype=theano.config.floatX)

        W_values = np.asarray(
            initWeights(self.filter_shape,
                        scale,
                        mode,
                        pool=self.pool),
            dtype=theano.config.floatX)

        self.W.set_value(W_values)
        self.b.set_value(b_values)

    def NLL(self,
            y,
            class_weights=None,
            example_weights=None,
            mask_class_labeled=None,
            mask_class_not_present=None,
            label_prop_thresh=None):
        """
        Returns the symbolic mean and instance-wise negative log-likelihood of the prediction
        of this model under a given target distribution.

        y: theano.tensor.TensorType
          corresponds to a vector that gives for each example the correct label. Labels < 0 are ignored (e.g. can
          be used for label propagation)

        class_weights: theano.tensor.TensorType
          weight vector of float32 of length  ``n_lab``. Values: ``1.0`` (default), ``w < 1.0`` (less important),
          ``w > 1.0`` (more important class)

        example_weights: theano.tensor.TensorType
          weight vector of float32 of shape ``(bs, z, x, y) that can give the individual examples (i.e. labels for
          output pixels) different weights. Values: ``1.0`` (default), ``w < 1.0`` (less important),
          ``w > 1.0`` (more important example). Note: if this is not normalised/bounded it may result in a
          effectively modified learning rate!

        The following refers to lazy labels, the masks are always on a per patch basis, depending on the
        origin cube of the patch. The masks are properties of the individual image cubes and must be loaded
        into CNNData.

        mask_class_labeled: theano.tensor.TensorType
          shape = (batchsize, num_classes).
          Binary masks indicating whether a class is properly labeled in ``y``. If a class ``k``
          is (in general) present in the image patches **and** ``mask_class_labeled[k]==1``, then
          the labels  **must** obey ``y==k`` for all pixels where the class is present.
          If a class ``k`` is present in the image, but was not labeled (-> cheaper labels), set
          ``mask_class_labeled[k]=0``. Then all pixels for which the ``y==k`` will be ignored.
          Alternative: set ``y=-1`` to ignore those pixels.
          Limit case: ``mask_class_labeled[:]==1`` will result in the ordinary NLL.

        mask_class_not_present: theano.tensor.TensorType
          shape = (batchsize, num_classes).
          Binary mask indicating whether a class is present in the image patches.
          ``mask_class_not_present[k]==1`` means that the image does **not** contain examples of class ``k``.
          Then for all pixels in the patch, class ``k`` predictive probabilities are trained towards ``0``.
          Limit case: ``mask_class_not_present[:]==0`` will result in the ordinary NLL.

        label_prop_thresh: float (0.5,1)
          This threshold allows unsupervised label propagation (only for examples with negative/ignore labels).
          If the predictive probability of the most likely class exceeds the threshold, this class is assumed to
          be the correct label and the training is pushed in this direction.
          Should only be used with pre-trained networks, and values <= 0.5 are disabled.

        Examples:

        - A cube contains no class ``k``. Instead of labelling the remaining classes they can be
          marked as unlabelled by the first mask (``mask_class_labeled[:]==0``, whether ``mask_class_labeled[k]``
          is ``0`` or ``1`` is actually indifferent because the labels should not be ``y==k`` anyway in this case).
          Additionally ``mask_class_not_present[k]==1`` (otherwise ``0``) to suppress predictions of ``k`` in
          in this patch. The actual value of the labels is indifferent, it can either be ``-1`` or it could be the
          background class, if the background is marked as unlabelled (i.e. then those labels are ignored).

        - Only part of the cube is densely labelled. Set ``mask_class_labeled[:]=1`` for all classes, but set the
          label values in the unlabelled part to ``-1`` to ignore this part.

        - Only a particular class ``k`` is labelled in the cube. Either set all other label pixels to ``-1`` or the
          corresponding flags in ``mask_class_labeled`` for the unlabelled classes.

        ..  Note::
          Using ``-1`` labels or telling that a class is not labelled, is somewhat redundant and just
          supported for convenience.
        """

        # NOTE: This whole function has a ugly problem with NaN. They arise for pred values close to 0 or 1
        # (i.e. for NNs that make very confident and usually also correct predictions) because initially the log of
        # all the whole pred tensor is taken. Later we want to use only some indices of the tensor (mask) but
        # that is not so easy done. There are two ways:
        # 1. advanced indexing: T.log(pred)[mask.nonzero()] --> fails if mask is all zero, cannot be fixed
        # 2. mutiplying with 0-1-mask: T.log(pred) * mask.nonzero --> but NaN * 0 = NaN, but we require 0!
        # For the second option, in principle, the NaNs could be replaced by 0 using T.switch, but then the gradient
        # fails because the replaced value is disconnected from the parameters and gives NaN (mathematically
        # the gradient should correctly be 0 then; there is a Theano ticket open to request a fix).
        # So finally the best practice is to add a stabilisation to the log: T.log(pred) --> T.log(pred+eps)
        # This looks ugly, but does the task and the introduced error is completely negligible.

        eps = 1e-6
        pred = self.class_probabilities  # predictive (bs, cl, x, y)
        y = y.dimshuffle(0, 'x', 1, 2, 3)  # the labels (bs, 1,  x, y)
        cls = T.arange(self.class_probabilities.shape[1]).dimshuffle('x', 0, 'x', 'x', 'x')  # available classes
        label_selection = T.eq(cls, y)  # selects correct labels

        if class_weights is None:
            class_weights = T.ones_like(pred)
        else:
            class_weights = class_weights.dimshuffle('x', 0, 'x', 'x', 'x')

        if example_weights is None:
            example_weights = T.ones_like(pred)
        else:
            example_weights = example_weights.dimshuffle(0, 'x', 1, 2, 3)

            # Up vote block
        if mask_class_labeled is not None:  # Standard
            mask_class_labeled = mask_class_labeled.dimshuffle(0, 1, 'x', 'x', 'x')
            label_selection = (label_selection * mask_class_labeled)

        nll_inst_up = -T.log(pred + eps) * label_selection * class_weights * example_weights
        N_up = T.sum(label_selection)  # number of labelled examples

        if label_prop_thresh is not None:  # Label propagation block
            above_thresh = pred > label_prop_thresh  # this is one for the class with highes prob
            prop_mask = above_thresh * (1 - mask_class_labeled)  # don't do where training labels are available
            nll_inst_up_prop = -T.log(pred + eps) * prop_mask * class_weights * example_weights
            N_up_prop = prop_mask.sum()

            nll_inst_up += nll_inst_up_prop
            N_up += N_up_prop

        if mask_class_not_present is not None:  # Down vote block
            mask_class_not_present = mask_class_not_present.dimshuffle(0, 1, 'x', 'x', 'x') * T.ones_like(y)
            nll_inst_dn = -T.log(1.0 - (pred * mask_class_not_present * class_weights * example_weights) + eps)
            N_dn = mask_class_not_present.sum()  # number of not present classes examples
        else:
            nll_inst_dn = 0.0
            N_dn = 0.0

            #nll_inst = (nll_inst_up + nll_inst_dn).sum(axis=1)
            #nll_inst = T.concatenate([nll_inst_up, nll_inst_dn, pred], axis=1)
        N_total = (N_up + N_dn)
        # patch N_total to be not 0, when this is the case the sum is 0 anyway:
        N_total = T.switch(T.eq(N_total, 0), 1, N_total)
        nll = (nll_inst_up + nll_inst_dn).sum() / N_total

        return nll, nll

    def NLL_weak(self,
                 y,
                 class_weights=None,
                 mask_class_labeled=None,
                 mask_class_not_present=None,
                 label_prop_thresh=None):
        """
        NLL that mixes the current cnn output and the hard labels as target
        """
        eps = 1e-6
        pred = self.class_probabilities  # predictive (bs, cl, x, y)
        y = y.dimshuffle(0, 'x', 1, 2, 3)  # the labels (bs, 1,  x, y)
        cls = T.arange(self.class_probabilities.shape[1]).dimshuffle('x', 0, 'x', 'x', 'x')  # available classes
        hard_labels = T.eq(cls, y)  # selects correct labels

        soft_labels = 0.5 * hard_labels + 0.5 * pred

        if class_weights is None:
            class_weights = T.ones_like(pred)
        else:
            class_weights = class_weights.dimshuffle('x', 0, 'x', 'x', 'x')

        nll_inst_up = -(soft_labels * T.log(pred + eps)) * class_weights
        #nll_inst_up = - 0.5 * (hard_labels * T.log(pred + eps)) - 0.5 * (pred * T.log(pred + eps))
        #nll_inst_up = -T.log(pred + eps)  * class_weights
        N_up = T.sum(hard_labels)  # number of labelled examples

        nll_inst_dn = 0.0
        N_dn = 0.0

        N_total = (N_up + N_dn)
        # patch N_total to be not 0, when this is the case the sum is 0 anyway:
        N_total = T.switch(T.eq(N_total, 0), 1, N_total)
        nll = (nll_inst_up + nll_inst_dn).sum() / N_total

        return nll, nll

    def NLL_affinity(self,
                     y,
                     class_weights=None,
                     mask_class_labeled=None,
                     mask_class_not_present=None,
                     label_prop_thresh=None):
        """
        TODO
        """
        pred = self.class_probabilities  # predictive (bs,3,z,x,y)
        y = y  # the labels  (bs, 3, z, x, y)

        if class_weights is None:
            class_weights = [1.0, 1.0]

        #nll = -(y * T.log(pred) * class_weights[1] + (1.0 - y) * T.log(1.0 - pred)) * class_weights[0]
        nll = -T.xlogx.xlogy0(y, pred + 1e-8) * class_weights[1] \
            +  T.xlogx.xlogy0((1.0 - y), (1.0 - pred + 1e-8)) * class_weights[0]

        nll_inst = nll
        nll = nll.mean()

        return nll, nll_inst

    def squared_distance(self, y):
        """
        Returns squared distance between prediction and ``y``

        :type y: theano.tensor.TensorType
        :param y: corresponds to a vector that gives for each example the
                      correct label
        """
        return T.mean((self.output - y)**2)

    def errors(self, y):
        """
        Returns classification accuracy

        :type y: theano.tensor.TensorType
        :param y: corresponds to a vector that gives for each example the
                  correct label
        """
        # check if y has same dimension of y_pred
        if y.ndim != self.class_prediction.ndim:
            raise TypeError('y should have the same shape as self.class_prediction',
                            ('y', y.type, 'class_prediction', self.class_prediction.type))
        # check if y is of the correct datatype
        if y.dtype.startswith('int'):
            return T.mean(T.neq(self.class_prediction, y)[T.ge(y, 0).nonzero()])
        else:
            print "something went wrong"
            raise NotImplementedError()


class AffinityLayer3d(object):
    def __init__(self,
                 input,
                 input_shape,
                 filter_shape,
                 pool,
                 activation_func,
                 enable_dropout,
                 use_fragment_pooling,
                 reshape_output,
                 mfp_offsets,
                 mfp_strides,
                 input_layer=None,
                 W=None,
                 b=None,
                 pooling_mode='max'):

        W_values1 = np.asarray(
            initWeights(filter_shape,
                        scale='glorot',
                        mode='normal',
                        pool=pool),
            dtype=theano.config.floatX)
        W_values2 = np.asarray(
            initWeights(filter_shape,
                        scale='glorot',
                        mode='normal',
                        pool=pool),
            dtype=theano.config.floatX)
        W_values3 = np.asarray(
            initWeights(filter_shape,
                        scale='glorot',
                        mode='normal',
                        pool=pool),
            dtype=theano.config.floatX)

        W_values = np.concatenate([W_values1, W_values2, W_values3], axis=0)
        self.W = theano.shared(W_values, name='W_conv', borrow=True)

        # the bias is a 1D tensor -- one bias per output feature map
        if activation_func in ['ReLU', 'relu']:
            norm = filter_shape[1] * filter_shape[3] * filter_shape[4]
            b_values = np.ones((filter_shape[0], ), dtype=theano.config.floatX) / norm
        if b is None:
            n_out = filter_shape[0]
            if activation_func == 'relu' or activation_func == 'ReLU':
                norm = filter_shape[1] * filter_shape[3] * filter_shape[4]
                b_values = np.asarray(
                    initWeights(
                        (n_out, ),
                        scale=1.0 / norm,
                        mode='const'),
                    dtype=theano.config.floatX)
            elif activation_func == 'sigmoid':
                b_values = np.asarray(
                    initWeights(
                        (n_out, ),
                        scale=0.5,
                        mode='const'),
                    dtype=theano.config.floatX)
            else:  # activation_func=='tanh':
                b_values = np.asarray(
                    initWeights(
                        (n_out, ),
                        scale=1e-6,
                        mode='fix-uni'),
                    dtype=theano.config.floatX)

            b_values = np.concatenate([b_values, b_values, b_values], axis=0)
            self.b = theano.shared(value=b_values, borrow=True, name='b_conv')

        N = filter_shape[0]

        l1 = ConvLayer3d(input,
                         input_shape,
                         filter_shape,
                         pool,
                         activation_func,
                         enable_dropout,
                         use_fragment_pooling,
                         reshape_output,
                         mfp_offsets,
                         mfp_strides,
                         affinity=False,
                         W=self.W[0:N],
                         b=self.b[0:N])
        l2 = ConvLayer3d(input,
                         input_shape,
                         filter_shape,
                         pool,
                         activation_func,
                         enable_dropout,
                         use_fragment_pooling,
                         reshape_output,
                         mfp_offsets,
                         mfp_strides,
                         affinity=False,
                         W=self.W[N:2 * N],
                         b=self.b[N:2 * N])
        l3 = ConvLayer3d(input,
                         input_shape,
                         filter_shape,
                         pool,
                         activation_func,
                         enable_dropout,
                         use_fragment_pooling,
                         reshape_output,
                         mfp_offsets,
                         mfp_strides,
                         affinity=False,
                         W=self.W[2 * N:3 * N],
                         b=self.b[2 * N:3 * N])

        self.params = [self.W, self.b]
        self.mfp_strides = l1.mfp_strides
        self.mfp_offsets = l1.mfp_offsets
        self.pool = l1.pool
        self.output_shape = list(l1.output_shape)
        self.output_shape[2] = self.output_shape[2] * 3
        self.prob_shape = list(l1.prob_shape)
        self.prob_shape[1] = self.output_shape[1] * 3
        self.activation_func = l1.activation_func

        self.class_probabilities = T.concatenate(
            [l1.class_probabilities, l2.class_probabilities, l3.class_probabilities], axis=1)
        self.class_prediction = T.concatenate(
            [l1.class_prediction, l2.class_prediction, l3.class_prediction], axis=1)

        self.l1 = l1
        self.l2 = l2
        self.l3 = l3

    def NLL_affinity(self,
                     y,
                     class_weights=None,
                     example_weights=None,
                     mask_class_labeled=None,
                     mask_class_not_present=None,
                     label_prop_thresh=None):

        nll1, _ = self.l1.NLL(y[:, 0], class_weights, example_weights,
                              mask_class_labeled, mask_class_not_present, label_prop_thresh)
        nll2, _ = self.l2.NLL(y[:, 1], class_weights, example_weights,
                              mask_class_labeled, mask_class_not_present, label_prop_thresh)
        nll3, _ = self.l3.NLL(y[:, 2], class_weights, example_weights,
                              mask_class_labeled, mask_class_not_present, label_prop_thresh)

        nll = T.mean(T.stack([nll1, nll2, nll3]))

        return nll, nll

    def errors(self, y):
        err1 = self.l1.errors(y[:, 0])
        err2 = self.l2.errors(y[:, 1])
        err3 = self.l3.errors(y[:, 2])
        err = T.mean(T.stack([err1, err2, err3]))

        return err


class MalisLayer(AffinityLayer3d):
    def __init__(self,
                 input,
                 input_shape,
                 filter_shape,
                 pool,
                 activation_func,
                 enable_dropout,
                 use_fragment_pooling,
                 reshape_output,
                 mfp_offsets,
                 mfp_strides,
                 input_layer=None,
                 W=None,
                 b=None,
                 pooling_mode='max'):

        super(MalisLayer, self).__init__(
            input, input_shape, filter_shape, pool, activation_func,
            enable_dropout, use_fragment_pooling, reshape_output, mfp_offsets,
            mfp_strides, input_layer, W, b, pooling_mode)

    def NLL_Malis(self, aff_gt, seg_gt, unrestrict_neg=True):
        """
        Parameters
        ----------

        aff_gt: 4d, (bs, #edges, x, y, z) int16

        seg_gt: (bs, x, y, z) int16


        Returns
        -------

        pos_count: for every edge number of pixel-pairs that should be connected by this edge
                   (excluding background/ECS pixels and only edges considered within the same object,
                   such that paths that go out from an object and back to the same object are irgnored)
        neg_count: for every edge number of pixel-pairs that should be separated by this edge
                   (excluding background/ECS pixels and only edges considered between objects,
                   such that minimal edges inside an object are not consideres to play a role for separating objects)
        unrestrict_neg: Bool
                    Use this to relax the restriction on neg_counts. The restriction
                    modifies the edge weights for before calculating the negative counts
                    as: ``edge_weights_neg = np.maximum(affinity_pred, affinity_gt)``
                    If unrestricted the predictions are used directly.
        """
        n_class = self.filter_shape[0]
        # prob.shape = (bs, 6, z,x,y) 6--> edge1 neg, edge1 pos, edge2 neg...
        affinity_pred = self.class_probabilities[0, 1::n_class]
        disconnect_pred = self.class_probabilities[0, 0::n_class]
        neigh_pattern = np.eye(3, dtype=np.int32)
        aff_gt = aff_gt[0]  # strip batch dimension
        seg_gt = seg_gt[0]  # strip batch dimension

        pos_count, neg_count = malis_weights(affinity_pred, aff_gt, seg_gt, neigh_pattern, unrestrict_neg)

        n_pos = T.sum(pos_count)
        n_neg = T.sum(neg_count)
        n_tot = n_pos + n_neg

        weighted_pos = -T.xlogx.xlogy0(pos_count, affinity_pred)  # drive up prediction for "connected" here
        weighted_neg = -T.xlogx.xlogy0(neg_count, disconnect_pred)  # drive up prediction for "disconnected" here
        nll = T.sum(weighted_pos + weighted_neg) / (n_tot + 1e-6)

        false_splits = T.sum((affinity_pred < 0.5) * pos_count)
        false_merges = T.sum((affinity_pred > 0.5) * neg_count)

        rand_index = T.cast(false_splits + false_merges, 'float32') / (n_tot + 1e-6)

        # eg 0.0   5187779 4578211 9765990 3439497477 7732598379 1143.9798583984375)
        return nll, n_pos, n_neg, n_tot, false_splits, false_merges, rand_index, pos_count, neg_count


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
# ELEKTRONN - Neural Network Toolkit
#
# Copyright (c) 2014 - now
# Max-Planck-Institute for Medical Research, Heidelberg, Germany
# Authors: Marius Killinger, Gregor Urban

import numpy as np
import multiprocessing as mp
import ctypes
import logging
import time
from collections import deque


#----------------------------------------------------------------------------------------------
class SharedMem(object):
    """Utilities to share np.arrays between processes"""
    _ctypes_to_numpy = {
        ctypes.c_int8: np.dtype(np.int8),
        ctypes.c_uint8: np.dtype(np.uint8),
        ctypes.c_int16: np.dtype(np.int16),
        ctypes.c_uint16: np.dtype(np.uint16),
        ctypes.c_int32: np.dtype(np.int32),
        ctypes.c_uint32: np.dtype(np.uint32),
        ctypes.c_int64: np.dtype(np.int64),
        ctypes.c_uint64: np.dtype(np.uint64),
        ctypes.c_byte: np.dtype(np.int8),
        ctypes.c_ubyte: np.dtype(np.uint8),
        ctypes.c_short: np.dtype(np.int16),
        ctypes.c_ushort: np.dtype(np.uint16),
        ctypes.c_int: np.dtype(np.int32),
        ctypes.c_uint: np.dtype(np.uint32),
        ctypes.c_long: np.dtype(np.int32),
        ctypes.c_ulong: np.dtype(np.uint32),
        ctypes.c_longlong: np.dtype(np.int64),
        ctypes.c_ulonglong: np.dtype(np.int64),
        ctypes.c_float: np.dtype(np.float32),
        ctypes.c_double: np.dtype(np.float64)
    }

    _numpy_to_ctypes = dict(zip(_ctypes_to_numpy.values(),
                                _ctypes_to_numpy.keys()))

    @staticmethod
    def shm2ndarray(mp_array, shape=None):
        """
        Parameters
        ----------

        mp_array: a mp.Array
        shape:    (optional) the returned np.ndarray is reshaped to this shape, flat otherwise

        Returns
        -------

        array: np.ndarray
         That can be normally used but changes are reflected in shared mem

        Note: the returned array is still pointing to the sharedmem, data might be changed by another process!
        """

        #if not hasattr(mp_array, '_type_'):
        #  mp_array = mp_array.get_obj()

        dtype = SharedMem._ctypes_to_numpy[mp_array._type_]
        result = np.frombuffer(mp_array, dtype)

        if shape is not None:
            #assert np.prod(shape)==result.size, "Cannot reshape length-%s array to shape %s"%(result.size, shape)
            result = result.reshape(shape)

        return np.asarray(result)

    @staticmethod
    def ndarray2shm(np_array, lock=False):
        """
        Parameters
        ----------
        np_array: np.ndarray
          array of arbitrary shape
        lock: Bool
          Whether to create a multiprocessing.Lock
        Returns
        -------
        handle: mp.Array:
          flat with data from ndarray copied to it
        """
        array1d = np_array.ravel(order='A')

        try:
            c_type = SharedMem._numpy_to_ctypes[array1d.dtype]
        except KeyError:
            c_type = SharedMem._numpy_to_ctypes[np.dtype(array1d.dtype)]

        result = mp.Array(c_type, array1d.size, lock=lock)
        SharedMem.shm2ndarray(result)[:] = array1d
        return result

    def puthandle(self, dtype, shape, data=None, lock=False):
        """
        Creates new shared memory and puts it on the queue. Other sub-processes can write to it.

        Parameters
        ----------
        dtype: np.dtype
          Type of data to store in array
        shape: tuple
          Properties of shared mem to be created
        data: np.ndarray
         (optional) values to fill shared array with
        lock: Bool
          Whether to create a multiprocessing.Lock on the shared variable

        Returns
        -------
        sharedmem handle: mp.array
        """
        t0 = time.clock()
        size = np.prod(shape)
        try:
            c_type = SharedMem._numpy_to_ctypes[dtype]
        except KeyError:
            c_type = SharedMem._numpy_to_ctypes[np.dtype(dtype)]

        shm = mp.Array(c_type, size, lock=lock)
        t1 = time.clock()

        if data is not None:
            SharedMem.shm2ndarray(shm)[:] = data.astype(dtype).ravel(order='A')
        t2 = time.clock()

        if self.profile:
            t_alloc = t1 - t0
            t_write = t2 - t1
            self.logger.info('SharedMemAlloc %g ms, WriteInitialData %g ms' %
                             (t_alloc * 1000, t_write * 1000))

        return shm


class Proc(mp.Process):
    """
    A *reusable* and *configurable* background process, that does the same job every time
    ``events['new']`` is set and signals that is has finished one iteration by setting ``events['ready']``
    """

    def __init__(self, mp_arrays, shapes, events, target, target_args,
                 target_kwargs, profile):
        super(Proc, self).__init__()
        self.events = events
        self.target = target
        self.target_args = target_args
        self.target_kwargs = target_kwargs
        self.arrays = []  # shm "wrapped" as np.array-objs
        self.profile = profile

        if profile:
            self.logger = mp.log_to_stderr(logging.INFO)

        for shm, shp in zip(mp_arrays, shapes):
            self.arrays.append(SharedMem.shm2ndarray(shm, shp))

    def run(self):
        while True:
            try:
                self.events['new'].wait()  # wait till host has fetched data from shm and demand new data from this proc
                self.events['new'].clear()
                t0 = time.clock()
                result = self.target(*self.target_args, **self.target_kwargs)
                for a, r in zip(self.arrays, result):
                    a[:] = r

                self.events['ready'].set()  # signal host that task is done and data is ready in shm
                t1 = time.clock()
                if self.profile:
                    t_exec = t1 - t0
                    self.logger.info('Executing Target and writing to shm %g ms' % (t_exec * 1000))
            except KeyboardInterrupt:
                pass


class BackgroundProc(SharedMem):
    def __init__(self,
                 target,
                 dtypes=None,
                 shapes=None,
                 n_proc=1,
                 target_args=(),
                 target_kwargs={},
                 profile=False):
        """
        Data structure to manage repeated background tasks by reusing a fixed number of *initially* created
        background process with the same arguments at every time. (E.g. retrieving an augmented batch)
        Remember to call ``BackgroundProc.shutdown`` after use to avoid zombie process and RAM clutter.

        Parameters
        ----------

        dtypes:
          list of dtypes of the target return values
        shapes:
          list of shapes of the target return values
        n_proc: int
          number of background procs to use
        target: callable
          target function for background proc. Can even be a method of an object, if object\
        data is read-only (then data will not be copied in RAM and the new process is lean). If\
        several procs use random modules, new seeds must be created inside target because they\
        have the same random state at the beginning.
        target_args:  tuple
          Proc args (constant)
        target_kwargs: dict
          Proc kwargs (constant)
        profile: Bool
          Whether to print timing results in to stdout

        Examples
        --------

        Use case to retrieve batches from a data structure ``D``:

          >>> data, label = D.getbatch(2, strided=False,
          flip=True, grey_augment_channels=[0])
          >>> kwargs = {'strided': False, 'flip': True, 'grey_augment_channels': [0]}
          >>> bg = BackgroundProc([np.float32, np.int16], [data.shape,label.shape],
          D.getbatch,n_proc=2, target_args=(2,), target_kwargs=kwargs, profile=False)
          >>> for i in xrange(100):
          >>>   data, label = bg.get()

        """
        self.dtypes = dtypes
        self.shapes = shapes
        self.target = target
        self.n_proc = n_proc
        self.i = 0  # index of next item to consume
        self.mp_arrays = []
        self.procs = []
        self.events = []
        self.profile = profile

        if (dtypes is None) or (shapes is None):
            ret = target(*target_args, **target_kwargs)
            dtypes = [b.dtype for b in ret]
            shapes = [b.shape for b in ret]
            self.dtypes = dtypes
            self.shapes = shapes

        if profile:
            self.logger = mp.log_to_stderr(logging.INFO)

        for k in xrange(n_proc):  # create a list of mp-arrays for each process
            a = []
            for dtype, shape in zip(dtypes, shapes):
                a.append(self.puthandle(dtype, shape))

            self.mp_arrays.append(a)
            self.events.append({'new': mp.Event(), 'ready': mp.Event()})

        for shm, e in zip(self.mp_arrays, self.events):  # initialise the procs and give them their mp-arrays
            p = Proc(shm, shapes, e, target, target_args, target_kwargs, profile)
            p.start()
            e['new'].set()
            self.procs.append(p)

    def get(self):
        """
        This gets the next result from a background process and blocks until the corresponding proc
        has finished.
        """
        k = self.i
        self.i = (self.i + 1) % self.n_proc  # advance index of next item
        result = []
        t0 = time.clock()
        self.events[k]['ready'].wait()
        self.events[k]['ready'].clear()
        t1 = time.clock()
        for shm, shp in zip(self.mp_arrays[k], self.shapes):
            result.append(SharedMem.shm2ndarray(shm, shp).copy())  # copy! Otherwise a proc will write to result

        self.events[k]['new'].set()
        t2 = time.clock()
        if self.profile:
            t_wait = t1 - t0
            t_write = t2 - t1
            self.logger.info(
                'Waiting for subprocess %g ms, converting to numpy %g ms' % (t_wait * 1000, t_write * 1000))

        return tuple(result)

    def shutdown(self):
        """**Must be called to free memory** if the background tasks are no longer needed"""
        for p in self.procs:
            p.terminate()

    def reset(self):
        """
        Should be called after an exception (e.g. by pressing ctrl+c) was raised.
        """
        for e in self.events:
            e['new'].set()


class SharedQ(SharedMem):
    """
    FIFO Queue to process np.ndarrays in the background (also pre-loading of data from disk)

    procs must accept list of ``mp.Array`` and make items ``np.ndarray`` using ``SharedQ.shm2ndarray``,\
    for this the shapes are required as too. The target requires the signature::

       >>> target(mp_arrays, shapes, *args, **kwargs)

    Whereas mp_array and shape are *automatically* added internally

    All parameters are optional:

    Parameters
    ----------


    n_proc: int
      If larger than 0, a message is printed if to few processes are running
    default_target: callable
      Default background proc callable
    default_args: tuple
      Default background proc and their parameters
    default_kwargs: dict
      Default background proc kwargs
    profile: Bool
      Whether to print timing results in terminal

    Examples
    ---------

    Automatic use:

      >>> Q = SharedQ(n_proc=2)
      >>> Q.startproc(target=, shape= args=, kwargs=)
      >>> Q.startproc(target=, shape= args=, kwargs=)
      >>> for i in xrange(5):
      >>>   Q.startproc(target=, shape= args=, kwargs=)
      >>>   item = Q.get() # starts as many new jobs as to maintain n_proc
      >>>   dosomehtingelse(item) # processes work in background to pre-fetch data for next iteration

    """

    def __init__(self,
                 n_proc=0,
                 default_target=None,
                 default_args=(),
                 default_kwargs={},
                 profile=False):

        self.data = deque()  # items of type [shm, shape, proc]
        self.len = 0
        self.n_proc = n_proc
        self.default_target = default_target
        self.default_args = default_args
        self.default_kwargs = default_kwargs
        self.profile = profile

    def startproc(self,
                  dtypes,
                  shapes,
                  target=None,
                  target_args=(),
                  target_kwargs={}):
        """
        Starts a new process

        procs must accept list of ``mp.Array`` and make items ``np.ndarray`` using ``SharedQ.shm2ndarray``,\
        for this the shapes are required as too. The target requires the signature::

           target(mp_arrays, shapes, *args, **kwargs)

        Whereas mp_array  and shape are *automatically* added internally
        """

        data = target_kwargs.get('data')

        if target is None:
            target = self.default_target
        if target_args == ():
            target_args = self.default_args
        if target_kwargs == {}:
            target_kwargs = self.default_kwargs

        mp_arrays = []
        for dtype, shape in zip(dtypes, shapes):
            mp_arrays.append(self.puthandle(dtype, shape, data))

        t0 = time.clock()
        _args = (mp_arrays, shapes) + target_args
        proc = mp.Process(target=target, args=_args, kwargs=target_kwargs)
        proc.start()

        self.data.append([mp_arrays, shapes, proc])
        self.len += 1
        t1 = time.clock()
        if self.profile:
            t_start = t1 - t0
            self.logger.info('Start Process %g ms' % (t_start * 1000))

    def get(self):
        """
        This gets the first results in the queue and blocks until the corresponding proc
        has finished. If a n_proc value is defined this then new procs must be started *before* to
        avoid a warning message.
        """
        mp_arrays, shapes, proc = self.data.popleft()
        self.len -= 1
        missing = self.n_proc - self.len
        if missing > 0:
            print "WARNING: You should have started %i new workes before Q.get()" % missing

        t0 = time.clock()
        proc.join()
        t1 = time.clock()
        result = []
        for shm, shp in zip(mp_arrays, shapes):
            result.append(SharedMem.shm2ndarray(shm, shp))

        t2 = time.clock()

        if self.profile:
            t_join = t1 - t0
            t_conv = t2 - t1
            self.logger.info('Join %g ms, Shared2Numpy %g ms' % (t_join * 1000, t_conv * 1000))

        return result

### Testing etc. ##############################################################################
# Pre requisits
if __name__ == "__main__":
    import gc
    import h5py

    def load():
        f = h5py.File('~/devel/data/MPI/raw_center_cube_mag1_v3.h5', 'r')
        d = f['raw'].value
        f.close()
        return d[0]

    t0 = time.time()
    D = load()
    t1 = time.time()
    lt = t1 - t0
    print('REAL LOAD TIME %.2f  sec' % (lt))

    def CPU():
        a = np.random.rand(1160 * 480)
        for i in xrange(50):
            np.sin(a)

    t0 = time.time()
    for i in xrange(3):
        CPU()

    t1 = time.time()
    rt = (t1 - t0) / 3
    print('REAL CPU TASK TIME %.2f  sec' % (rt))
    serial = 20 * rt + 20 * lt
    D = None
    gc.collect()

    def IO(mp_array, shape):
        t0 = time.time()
        d = load()
        SharedQ.shm2ndarray(mp_array, shape)[:] = d
        t = time.time() - t0
        print('LOADED data in %.2f  sec' % (t))

# Automated Process Approach ################################################################
#if False:
#  bg = BackgroundProc([np.uint8], [(1160, 480)], load, n_proc=2)
#  t00 = time.time()
#  for i in xrange(20):
#    t0 = time.time()
#    d = bg.get()
#    t2 = time.time()
#    t  = t2 - t0
#  #  logger.info('Start, Join, Popping   %.2f secs' %t)
#    CPU()
#    t4 = time.time()
#    t  = t4 - t2
#    print('CPU task  %.2f secs' %t)
#
#  t11 = time.time()
#  total = (t11-t00)
#  print('True time %.2f  sec, serial estimate %.2f' %(total, serial))
#  bg.shutdown()
#
#if False:
#  Q = SharedQ(n_proc=1, dtype=np.uint8, shape=(1160, 480), default_target=IO, profile=False)
#
#  t00 = time.time()
#  Q.startproc()
#  for i in xrange(20):
#    t0 = time.time()
#    d = Q.get()
#    t2 = time.time()
#    t  = t2 - t0
#  #  logger.info('Start, Join, Popping   %.2f secs' %t)
#    CPU()
#    t4 = time.time()
#    t  = t4 - t2
#    print('CPU task  %.2f secs' %t)
#
#  t11 = time.time()
#  total = (t11-t00)
#  print('True time %.2f  sec, serial estimate %.2f' %(total, serial))
#
#  print '\n\n\n\n'

# Process Approach ##########################################################################
#if False:
#  Q = SharedQ(0, np.uint8, (1160, 480))
#
#  t00 = time.time()
#  mp_array, shape, _ = Q.puthandle()
#  loader = mp.Process(target=IO, args=(mp_array, shape)) # start loading first item
#  loader.start()
#  for i in xrange(5):
#    t0 = time.time()
#
#    loader.join()
#    t1 = time.time()
#    t  = t1 - t0
#    print('Join Wait %.2f secs' %t)
#
#    d = Q.get()
#    t2 = time.time()
#    t  = t2 - t1
#  #  logger.info('Popping   %.2f secs' %t)
#
#    mp_array, shape, _ = Q.puthandle()
#    loader = mp.Process(target=IO, args=(mp_array, shape))
#    loader.start()
#    t3 = time.time()
#    t  = t3 - t1
#    print('Starting  %.2f secs' %t)
#
#    #CPUtask(d)
#    CPU()
#    t4 = time.time()
#    t  = t4 - t3
#    print('CPU task  %.2f secs' %t)
#
#  t11 = time.time()
#  total = (t11-t00)
#  print('True time %.2f  sec, serial estimate %.2f' %(total, serial))

# [INFO/MainProcess] True time 5.95  sec, serial estimate 8.77

# Threaded Approach #########################################################################
#DataQ = deque()
#
#def IOtask():
#  t0 =  time.time()
#  d = load()
#  DataQ.append(d)
#  t = time.time() - t0
#  logging.info('LOADED data in %.2f  sec' %(t))
#
#class IOproc(th.Thread):
#  def run(self):
#    time.sleep(0.000001)
#    t0 =  time.time()
#    d = load()
#
#    DataQ.append(d)
#    t = time.time() - t0
#    logging.info('LOADED data in %.2f  sec' %(t))
#
#class CPUproc(th.Thread):
#  def run(self):
#    time.sleep(0.000001)
#    t0 =  time.time()
#    a = np.random.rand(1160*480)
#    for i in xrange(50):
#      np.sin(a)
#    t = time.time() - t0
#    logging.info('LOADED data in %.2f  sec' %(t))
#
#t00 = time.time()
#loader = IOproc() # start loading first item
#loader.start()
#for i in xrange(5):
#  t0 = time.time()
#
#  loader.join()
#  t1 = time.time()
#  t  = t1 - t0
#  logging.info('Join Wait %.2f secs' %t)
#
#  d = DataQ.popleft()
#  t2 = time.time()
#  t  = t2 - t1
#  logging.info('Popping   %.2f secs' %t)
#
#  #loader = th.Thread(target=IOtask, args=())
#  loader = IOproc()
#  loader.start()
#  t3 = time.time()
#  t  = t3 - t1
#  logging.info('Starting  %.2f secs' %t)
#
#  #CPUtask(d)
#  cpu = CPUproc()
#  cpu.start()
#  cpu.join()
#  t4 = time.time()
#  t  = t4 - t3
#  logging.info('CPU task  %.2f secs' %t)
#
#t11 = time.time()
#total = (t11-t00)
#serial = 5*rt + 5*lt
#logging.info('True time %.2f  sec, serial estimate %.2f' %(total, serial))

# [INFO] (MainThread) True time 10.28  sec, serial estimate 8.68

# Serial Approach #########################################################################
#t00 = time.time()
#for i in xrange(5):
#  t0 = time.time()
#  d  = load()
#  t1 = time.time()
#  t  = t1 - t0
#  logging.info('Loading Wait %.2f secs' %t)
#
#  CPUtask(d)
#  t4 = time.time()
#  t  = t4 - t1
#  logging.info('CPU task  %.2f secs' %t)
#
#t11 = time.time()
#total = (t11-t00)
#serial = 5*rt + 5*lt
#logging.info('True time %.2f  sec, serial estimate %.2f' %(total, serial))

# [INFO] (MainThread) True time 8.93  sec, serial estimate 8.75



# -*- coding: utf-8 -*-
# ELEKTRONN - Neural Network Toolkit
#
# Copyright (c) 2014 - now
# Max-Planck-Institute for Medical Research, Heidelberg, Germany
# Authors: Marius Killinger, Gregor Urban

import sys
sys.setrecursionlimit(5000)  # Needed to compile MFP for large CNNs
